from oslo_log import log as logging
from wsme.rest.json import fromjson as to_model, tojson as from_model
from wsme.exc import InvalidInput
from pecan.rest import RestController
from pecan import request, response

from magellan.api.controllers.v1.resources import ExclusionList
from magellan.api.controllers.v1.usertypes import GUID
from magellan.utils.wsme_utils import expose

LOG = logging.getLogger(__name__)


class ExclusionListController(RestController):

    _custom_actions = {
        'default': ['GET']
    }

    def __init__(self, rpc_client):
        super().__init__()
        self.rpc_client = rpc_client

    @expose(ExclusionList, GUID())
    def get(self, exclusion_list_id):
        """
        Returns a ExclusionList identified by UUID.
        :param exclusion_list_id: UUID of exclusion list to retrieve
        """
        LOG.info("Fetch exclusion list [UUID=%s]", exclusion_list_id)
        exclusion_list = self.rpc_client.exclusion_list_get(exclusion_list_id)
        return to_model(ExclusionList, exclusion_list)

    @expose(ExclusionList)
    def default(self):
        """
        Retrieves or creates the default exclusion list.

        This provides a single default exclusion list that can be used until
        support for multiple exclusion lists is implemented in the frontend.
        """
        LOG.info("Fetch default exclusion list")
        exclusion_list = self.rpc_client.exclusion_list_get_default()
        return to_model(ExclusionList, exclusion_list)

    @expose(ExclusionList, GUID(), body=ExclusionList)
    def put(self, exclusion_list_id, exclusion_list_resource):
        """
        Update a exclusion list.
        :param exclusion_list_id: UUID of exclusion list to retrieve
        :param exclusion_list_resource: ExclusionList object to apply
        :return: The new exclusion list resource created
        """
        LOG.info("Update exclusion list [%s]", str(exclusion_list_resource))
        exclusion_list = self.rpc_client.exclusion_list_update(
            exclusion_list_id,
            from_model(ExclusionList, exclusion_list_resource))

        response.headers["Location"] = "%s/%s" % (
            request.url, exclusion_list["id"])
        return to_model(ExclusionList, exclusion_list)

    @expose(ExclusionList, body=ExclusionList, status_code=201)
    def post(self, exclusion_list_resource):
        """
        Create a new exclusion list.
        :param exclusion_list_resource: The new exclusion list
        resource to be created
        :return: The new exclusion list resource created
        """
        LOG.info("Create exclusion list [%s]", str(exclusion_list_resource))
        exclusion_list = self.rpc_client.exclusion_list_create(
            from_model(ExclusionList, exclusion_list_resource))
        response.headers["Location"] = "%s/%s" % (
            request.url, exclusion_list["id"])
        return to_model(ExclusionList, exclusion_list)

    @expose(None, GUID(), status_code=204)
    def delete(self, exclusion_list_id):
        """
        Delete the exclusion list.
        :param exclusion_list_id: UUID of exclusion list to delete
        """
        LOG.info("Delete exclusion list [UUID=%s]", exclusion_list_id)
        self.rpc_client.exclusion_list_delete(exclusion_list_id)

    @expose([ExclusionList])
    def get_all(self):
        """
        Return a list of all exclusion lists.
        """
        LOG.info("Fetch all exclusion lists.")
        exclusion_lists = self.rpc_client.exclusion_list_get_all()

        result = []
        for exclusion_list in exclusion_lists:
            try:
                result.append(to_model(ExclusionList, exclusion_list))
            except (ValueError, InvalidInput):
                LOG.error(
                    'Invalid exclusion list in database with id %s',
                    exclusion_list.id)
        return result
