import jsonpatch

from pecan.rest import RestController
from pecan import request, response
from oslo_log import log as logging
from wsme.rest.json import fromjson as to_model, tojson as from_model

from collections import defaultdict
from magellan.db import api as db_api
from magellan.common.consts import ExecutionStatus
from magellan.common.exceptions import InputException

from magellan.api.controllers.v1.resources import Device, Execution
from magellan.api.controllers.v1.usertypes import GUID, JsonType
from magellan.utils.wsme_utils import expose

LOG = logging.getLogger(__name__)


class DevicesController(RestController):

    def __init__(self, rpc_client):
        super().__init__()
        self.rpc_client = rpc_client

    @expose([Device], GUID())
    def get_all(self, execution_id):
        """Return a list of autodiscovery devices."""

        LOG.info("Fetch all devices for execution %s." % execution_id)
        devices = self.rpc_client.device_get_all(execution_id)
        connections = self.rpc_client.connection_get_all(execution_id)

        id_address_map = {}
        for device in devices:
            id_address_map[device.id] = device.address

        address_to_neighbour_map = defaultdict(set)
        for connection in connections:
            d1 = connection.device1_id
            d2 = connection.device2_id
            address_to_neighbour_map[d1].add(d2)
            address_to_neighbour_map[d2].add(d1)

        for device in devices:
            device.neighbours = list(map(id_address_map.get,
                                         address_to_neighbour_map[device.id]))

        return [to_model(Device, {
            "id": device.id,
            "address": device.address,
            "neighbours": device.neighbours,
            "details": device.details,
            "config": device.config,
            "status": device.status,
        }) for device in devices]

    @expose(Device, GUID(), GUID(), body=[JsonType()])
    def patch(self, execution_id, device_id, patch):
        """
        Update a device.
        :param execution_id: UUID of the execution the device belongs to
        :param device_id: UUID of the device to patch
        :param patch: JSON-PATCH object to apply
        :return: The updated device
        """
        LOG.info("Update device [UUID=%s]", device_id)
        old_device = self.rpc_client.device_get(device_id)
        try:
            patched = jsonpatch.apply_patch(dict(old_device), patch)
            # Re-validate as a Device after patching
            to_model(Device, {
                "id": patched["id"],
                "address": patched["address"],
                "details": patched["details"],
                "config": patched["config"],
                "status": patched["status"],
            })
        except (ValueError, jsonpatch.JsonPatchException) as e:
            raise InputException("Patch operation failed: %s" % str(e))
        new_device = self.rpc_client.device_update(device_id, patched)
        return to_model(Device, {
            "id": new_device.id,
            "address": new_device.address,
            "details": new_device.details,
            "config": new_device.config,
            "status": new_device.status,
        })


class ExecutionsController(RestController):

    _custom_actions = {
        'cancel': ['PUT']
    }

    def __init__(self, rpc_client):
        super().__init__()
        self.rpc_client = rpc_client
        self.devices = DevicesController(rpc_client)

    @expose(Execution, GUID())
    def get(self, execution_id):
        """Return the execution.
        :param execution_id: Id of execution to retrieve
        """
        LOG.info("Fetch execution [id=%s]", execution_id)
        db_model = self.rpc_client.execution_get(execution_id)
        return to_model(Execution, db_model)

    @expose(Execution, body=Execution, status_code=201)
    def post(self, new_resource):
        """Create a new execution."""

        LOG.info("Start new execution [%s]", str(new_resource))
        raw_execution = from_model(Execution, new_resource)

        raw_execution['status'] = ExecutionStatus.running
        execution_db = self.rpc_client.execution_create(raw_execution)
        response.headers["Location"] = "%s/%s" % (
            request.url,
            execution_db["id"]
        )

        return to_model(Execution, execution_db)

    @expose(None, GUID(), status_code=204)
    def delete(self, execution_id):
        """Delete the execution.
        :param execution_id: Name of execution to delete
        """
        LOG.info("Delete execution [id=%s]", execution_id)
        self.rpc_client.execution_delete(execution_id)

    @expose(Execution, GUID(), status_code=202)
    def cancel(self, execution_id):
        """Cancel an ongong discovery execution.
        :param execution_id: Name of execution to delete
        """
        LOG.info("Cancel execution [id=%s]", execution_id)
        db_model = self.rpc_client.execution_cancel(execution_id)
        return to_model(Execution, db_model)

    @expose([Execution])
    def get_all(self):
        """Return a list of autodiscovery executions."""
        LOG.info("Fetch all executions.")
        return [to_model(Execution, db_model) for db_model in
                db_api.execution_get_all()]

    @expose(Execution, GUID(), body=[JsonType()])
    def patch(self, execution_id, patch):
        """
        Update an execution.
        :param execution_id: UUID of the execution to patch
        :param patch: JSON-PATCH object to apply
        :return: The updated execution
        """
        LOG.info("Update execution [UUID=%s]", execution_id)
        old_execution = self.rpc_client.execution_get(execution_id)
        try:
            patched = jsonpatch.apply_patch(dict(old_execution), patch)
            # Re-validate as a Execution after patching
            to_model(Execution, patched)
        except (ValueError, jsonpatch.JsonPatchException) as e:
            raise InputException("Patch operation failed: %s" % str(e))
        new_execution = self.rpc_client.execution_update(execution_id, patched)
        return to_model(Execution, new_execution)
