import jsonpatch

from oslo_log import log as logging
from wsme.rest.json import fromjson as to_model, tojson as from_model
from wsme.exc import InvalidInput
from pecan.rest import RestController
from pecan import request, response

from magellan.api.controllers.v1.resources import Job, Execution
from magellan.api.controllers.v1.usertypes import GUID, JsonType
from magellan.common.exceptions import InputException
from magellan.utils.wsme_utils import expose

LOG = logging.getLogger(__name__)


class JobExecutionsController(RestController):

    def __init__(self, rpc_client):
        super().__init__()
        self.rpc_client = rpc_client

    @expose([Execution], GUID())
    def get_all(self, job_id):
        """Return a list of autodiscovery devices."""

        LOG.info("Fetch all executions for job %s." % job_id)
        executions = self.rpc_client.execution_get_all_by_job(job_id)

        return [to_model(Execution, execution) for execution in executions]


class JobsController(RestController):

    def __init__(self, rpc_client):
        super().__init__()
        self.rpc_client = rpc_client
        self.executions = JobExecutionsController(rpc_client)

    @expose(Job, GUID())
    def get(self, job_id):
        """
        Returns a Job identified by UUID.
        :param job_id: UUID of job to retrieve
        """
        LOG.info("Fetch job [UUID=%s]", job_id)
        job = self.rpc_client.job_get(job_id)
        executions = self.rpc_client.execution_get_all_by_job(job_id)
        if len(executions) > 0:
            setattr(job, "latest_execution", executions[0])
        return to_model(Job, job)

    @expose(Job, GUID(), body=[JsonType()])
    def patch(self, job_id, patch):
        """
        Update a job.
        :param job_id: UUID of job to retrieve
        :param patch: JSON-PATCH object to apply
        :return: The new job resource created
        """
        LOG.info("Update job [UUID=%s]", job_id)
        job_model = self.rpc_client.job_get(job_id)
        try:
            patched = jsonpatch.apply_patch(dict(job_model), patch)
            # Re-validate as a Job after patching
            to_model(Job, patched)
        except (ValueError, jsonpatch.JsonPatchException) as e:
            raise InputException("Patch operation failed: %s" % str(e))
        job = self.rpc_client.job_update(job_id, patched)
        return to_model(Job, job)

    @expose(Job, body=Job, status_code=201)
    def post(self, job_resource):
        """
        Create a new job.
        :param job_resource: The new job resource to be created
        :return: The new job resource created
        """
        LOG.info("Create job [%s]", str(job_resource))
        job = self.rpc_client.job_create(
            from_model(Job, job_resource))
        response.headers["Location"] = "%s/%s" % (request.url, job["id"])
        return to_model(Job, job)

    @expose(None, GUID(), status_code=204)
    def delete(self, job_id):
        """
        Delete the job.
        :param job_id: UUID of job to delete
        """
        LOG.info("Delete job [UUID=%s]", job_id)
        self.rpc_client.job_delete(job_id)

    @expose([Job])
    def get_all(self):
        """
        Return a list of autodiscovery jobs.
        """
        LOG.info("Fetch all jobs.")
        jobs = self.rpc_client.job_get_all()

        result = []
        for job in jobs:
            try:
                result.append(to_model(Job, job))
            except (ValueError, InvalidInput):
                LOG.error('Invalid job in database with id %s', job.id)
        return result
