from oslo_config import cfg
from wsme import types, wsattr, exc as wsexc
from wsme.rest.json import fromjson

from magellan.api.controllers.v1 import usertypes
from magellan.common import consts
from pyautodiscover.utils.IPAddress import validate_address_input


class SNMPv1(types.Base):
    version = usertypes.Constant(
        usertypes.Enum(consts.SnmpVersion), consts.SnmpVersion.v1)
    community = wsattr(types.text, default='public')

    @classmethod
    def sample(cls):
        return cls(version=consts.SnmpVersion.v1, community='public')


class SNMPv2(types.Base):
    version = usertypes.Constant(
        usertypes.Enum(consts.SnmpVersion), consts.SnmpVersion.v2c)
    community = wsattr(types.text, default='public')

    @classmethod
    def sample(cls):
        return cls(version=consts.SnmpVersion.v2c, community='public')


class SNMPv3AuthPriv(types.Base):
    version = usertypes.Constant(
        usertypes.Enum(consts.SnmpVersion), consts.SnmpVersion.v3
    )
    secname = wsattr(types.text, mandatory=True)
    authpass = wsattr(types.text, mandatory=True)
    authprotocol = wsattr(
        usertypes.ValueBasedEnum(consts.SnmpAuthProtocol),
        default=consts.SnmpAuthProtocol.SHA,
    )
    privpass = wsattr(types.text, mandatory=True)
    privprotocol = wsattr(
        usertypes.ValueBasedEnum(consts.SnmpPrivProtocol),
        default=consts.SnmpPrivProtocol.AES
    )

    @classmethod
    def sample(cls):
        return cls(
            version=consts.SnmpVersion.v3,
            secname='testing',
            authpass='testpass',
            privpass='testpass'
        )


class SNMPv3AuthNoPriv(types.Base):
    version = usertypes.Constant(
        usertypes.Enum(consts.SnmpVersion), consts.SnmpVersion.v3)
    secname = wsattr(types.text, mandatory=True)
    authpass = wsattr(types.text, mandatory=True)
    authprotocol = wsattr(
        usertypes.ValueBasedEnum(consts.SnmpAuthProtocol),
        default=consts.SnmpAuthProtocol.SHA
    )

    @classmethod
    def sample(cls):
        return cls(
            version=consts.SnmpVersion.v3,
            secname='testing',
            authpass='testpass'
        )


class SNMPv3NoAuthNoPriv(types.Base):
    version = usertypes.Constant(
        usertypes.Enum(consts.SnmpVersion), consts.SnmpVersion.v3)
    secname = wsattr(types.text, mandatory=True)

    @classmethod
    def sample(cls):
        return cls(
            version=consts.SnmpVersion.v3,
            secname='testing'
        )


class JobOptions(types.Base):
    """Valid job options."""

    snmp = usertypes.Union(
        'SNMP Options', SNMPv3NoAuthNoPriv,
        SNMPv3AuthNoPriv, SNMPv3AuthPriv, SNMPv2, SNMPv1)

    @classmethod
    def sample(cls):
        return cls(snmp=SNMPv2.sample())


class SeedRouterJobOptions(JobOptions):
    """Valid seed router job options."""

    seed_router = wsattr(usertypes.IPAddress(), mandatory=True)

    @classmethod
    def sample(cls):
        return cls(seed_router='127.0.0.1', snmp=SNMPv2.sample())


class IpRangeJobOptions(JobOptions):
    """Valid ip range scan job options."""

    ip_range = wsattr(usertypes.IPRange(), mandatory=True)

    @classmethod
    def sample(cls):
        return cls(ip_range='127.0.0.1-10', snmp=SNMPv2.sample())


class IpListJobOptions(JobOptions):
    """Valid ip list scan job options."""

    ip_list = wsattr(usertypes.IPList(), mandatory=False)

    @classmethod
    def sample(cls):
        return cls(
            ip_list='127.0.0.1 127.0.0.2 127.0.0.3 127.0.0.10-20',
            snmp=SNMPv2.sample()
        )


class Job(types.Base):
    """Job resource."""

    id = usertypes.GUID()

    name = wsattr(types.text, mandatory=True)
    description = wsattr(types.text, mandatory=True)
    method = wsattr(usertypes.Enum(consts.ScanMethod), mandatory=True)
    options = usertypes.Union(
        'Job Options', SeedRouterJobOptions,
        IpRangeJobOptions, IpListJobOptions
    )
    timeout = wsattr(int, default=cfg.CONF.discovery.timeout)
    schedule = wsattr(usertypes.Schedule(), default=None)
    latest_execution = wsattr('Execution', default=None)
    username = wsattr(types.text, default=None)

    created_at = usertypes.DateTime()
    updated_at = usertypes.DateTime()

    @staticmethod
    def validate(value):
        expected_options = {
            consts.ScanMethod.seed_router: SeedRouterJobOptions,
            consts.ScanMethod.ip_range: IpRangeJobOptions,
            consts.ScanMethod.ip_list: IpListJobOptions
        }
        if not isinstance(value.options, expected_options[value.method]):
            raise ValueError("Mismatching job method and options")
        return value

    @classmethod
    def sample(cls):
        return cls(
            id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440000'
            ),
            name='job',
            description='This is a long and useless'
            'description of an autodiscovery job',
            method=consts.ScanMethod.seed_router,
            options=SeedRouterJobOptions.sample(),
            timeout=3600,
            schedule=usertypes.Schedule().frombasetype("15 0 * * *"),
            created_at=usertypes.DateTime().frombasetype(
                '2017-01-01 00:00:00'
            ),
            updated_at=usertypes.DateTime().frombasetype(
                '2017-01-01 00:00:00'
            )
        )


class Execution(types.Base):
    """Execution resource."""

    id = usertypes.GUID()
    job_id = wsattr(usertypes.GUID(), mandatory=True)

    name = wsattr(types.text, mandatory=True)
    description = wsattr(types.text, mandatory=True)
    status = usertypes.Enum(consts.ExecutionStatus)
    schedule_status = usertypes.Enum(consts.ScheduleStatus)
    username = wsattr(types.text, default=None)

    complete_time = usertypes.DateTime()
    created_at = usertypes.DateTime()
    updated_at = usertypes.DateTime()

    @classmethod
    def sample(cls):
        return cls(
            id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440000'
            ),
            job_id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440001'
            ),
            name='execution',
            description='This is a long and useless'
            'description of an autodiscovery execution',
            status=consts.ExecutionStatus.canceled,
            schedule_status=consts.ScheduleStatus.unscheduled,
            complete_time=usertypes.DateTime().frombasetype(
                '2017-01-01 00:00:00'
            ),
            created_at=usertypes.DateTime().frombasetype(
                '2017-01-01 00:00:00'
            ),
            updated_at=usertypes.DateTime().frombasetype(
                '2017-01-01 00:00:00'
            )
        )


class ScheduledResults(Execution):
    """Execution resource with resulting unmonitored addresses."""

    unmonitored = [types.text]


class Device(types.Base):
    """Network device."""

    id = usertypes.GUID()
    address = types.text
    neighbours = [types.text]
    details = {
        str: str
    }
    config = {
        str: str
    }
    status = wsattr(usertypes.Enum(consts.DeviceStatus))

    @classmethod
    def sample(cls):
        return cls(
            id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440000'
            ),
            address='127.0.0.1',
            neighbours=['192.168.0.1', '192.168.0.2'],
            details={
                'sysName': 'MyRouter'
            }
        )


class ExclusionListItem(types.Base):
    """Exclusion list item resource."""

    id = usertypes.GUID()
    exclusion_list_id = wsattr(usertypes.GUID(), mandatory=False)
    address = wsattr(types.text, mandatory=True)
    comment = wsattr(types.text, mandatory=False)

    created_at = usertypes.DateTime()
    updated_at = usertypes.DateTime()

    @staticmethod
    def validate(value):
        if not value.address or not validate_address_input(value.address):
            raise ValueError("Invalid IP address, range or subnet")
        return value

    @classmethod
    def sample(cls):
        return cls(
            id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440000'
            ),
            exclusion_list_id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440001'
            ),
            address='192.168.1.1',
            comment='This is a comment'
        )


class ExclusionList(types.Base):
    """Exclusion list resource."""
    id = usertypes.GUID()
    name = wsattr(types.text, mandatory=True)
    description = wsattr(types.text, mandatory=False, default="")
    items = [ExclusionListItem]

    created_at = usertypes.DateTime()
    updated_at = usertypes.DateTime()

    @staticmethod
    def validate(value):
        if not value.name:
            raise ValueError("Empty name not allowed")
        if value.description is None:
            value.description = ""
        if value.items is None:
            value.items = []
        return value

    @classmethod
    def sample(cls):
        return cls(
            id=usertypes.GUID().frombasetype(
                '123e4567-e89b-12d3-a456-426655440000'
            ),
            name='127.0.0.1',
            description=['192.168.0.1', '192.168.0.2'],
            items=[]
        )


@fromjson.when_object(Job)
@fromjson.when_object(SeedRouterJobOptions)
@fromjson.when_object(IpRangeJobOptions)
@fromjson.when_object(IpListJobOptions)
@fromjson.when_object(Execution)
@fromjson.when_object(ExclusionList)
@fromjson.when_object(ExclusionListItem)
def try_fromjson(datatype, value):
    """
    This function adds a check when we convert input data to our defined WSME
    types. The generic fromjson implementation assumes that value is a
    dict-like object and does not handle the exception if it isn't.

    This means we get a server error if the user supplies a valid JSON
    datatype other than a JSON object for these types. We want that case to
    raise a InvalidInput exception instead.

    We can't directly check if value is a dict because it can also be a
    SqlAlchemy object (which behaves like a dict).

    It looks like this is a bug in WSME, but it could also be that we simply
    don't define our types correctly.
    """
    if value is None:
        return value
    if hasattr(value, 'keys'):
        return fromjson.default(datatype, value)
    raise wsexc.InvalidInput(datatype.__name__.lower(), value)
