'''
Interface for database access.

SQLAlchemy is currently the only supported backend.
'''

import uuid

from oslo_config import cfg
from oslo_db import api

from magellan.common.consts import ExecutionStatus, ScanMethod
from magellan.db.datafiles import read_ip_list, write_ip_list, remove_ip_list
from magellan.common.exceptions import (
    JobAlreadyRunningError,
    DBEntityNotFoundError,
)
from magellan.engine.discover import start_discovery, stop_discovery

CONF = cfg.CONF

_BACKEND_MAPPING = {'sqlalchemy': 'magellan.db.sqlalchemy.api'}

IMPL = api.DBAPI.from_config(CONF, backend_mapping=_BACKEND_MAPPING)


def get_engine():
    return IMPL.get_engine()


def get_session():
    return IMPL.get_session()


def db_sync(engine, version=None):
    """Migrate the database to `version` or the most recent version."""
    return IMPL.db_sync(engine, version=version)


def db_version(engine):
    """Display the current database version."""
    return IMPL.db_version(engine)


def job_create(values):
    """Create an autodiscovery job."""
    if values['method'] == ScanMethod.ip_list.name:
        ip_list = values['options'].pop('ip_list', None)
        job = IMPL.job_create(values)
        if ip_list:
            write_ip_list(job.id, ip_list)
    else:
        job = IMPL.job_create(values)
    return job


def job_get(job_id):
    """Retrieve an autodiscovery job or raise if it does not exist."""
    job = IMPL.job_get(job_id)
    if job.method == ScanMethod.ip_list:
        # For backwards compatibility check if there is ip_list data in the DB
        ip_list = job.options.pop('ip_list', None)
        if ip_list:
            write_ip_list(job_id, ip_list)
            job_update(job_id, dict(job))
        job.options['ip_list'] = read_ip_list(job_id)
    return job


def job_get_all():
    """Retrieve all registered jobs."""
    return IMPL.job_get_all()


def job_update(job_id, values):
    """Update a job or raise if it does not exist."""
    try:
        ip_list = values['options'].pop('ip_list')
    except KeyError:
        ip_list = None
    job = IMPL.job_update(job_id, values)
    if job.method != ScanMethod.ip_list:
        remove_ip_list(job_id)
    elif ip_list is not None:
        write_ip_list(job_id, ip_list)
        job.options['ip_list'] = ip_list
    return job


def job_delete(job_id):
    """Delete a job or raise if it does not exist."""
    IMPL.job_delete(job_id)
    remove_ip_list(job_id)


def execution_create(values):
    """Create an autodiscovery job."""
    # Make sure that an execution is not started while an execution is already
    # running for the same job
    job_executions = execution_get_all_by_job(values['job_id'])
    if len(job_executions) > 0:
        if job_executions[0].status == ExecutionStatus.running:
            raise JobAlreadyRunningError()

    job = job_get(values['job_id'])
    execution = IMPL.execution_create(values)
    exclusion_list = exclusion_list_get_default()
    start_discovery(execution, job, [
                    item.address for item in exclusion_list.items])
    return execution


def execution_get(execution_id):
    """Retrieve an autodiscovery execution or raise if it does not exist."""
    return IMPL.execution_get(execution_id)


def execution_get_all():
    """Retrieve all registered executions."""
    return IMPL.execution_get_all()


def execution_get_all_by_job(job_id):
    """Retrieve all registered executions for job."""
    return IMPL.execution_get_all_by_job(job_id)


def execution_get_all_scheduled(username):
    """Retrieve all latest scheduled executions for a user."""
    return IMPL.execution_get_all_scheduled(username)


def execution_cancel(execution_id):
    """Cancel an execution or raise if it does not exist."""
    execution = execution_get(execution_id)
    if stop_discovery(execution):
        execution = IMPL.execution_cancel(execution_id)
    return execution


def execution_update(execution_id, values):
    """Update an execution or raise if it does not exist."""
    return IMPL.execution_update(execution_id, values)


def execution_delete(execution_id):
    """Delete an execution or raise if it does not exist."""
    return IMPL.execution_delete(execution_id)


def device_get_all(execution_id):
    """Get all devices for an execution_id"""
    return IMPL.device_get_all(execution_id)


def device_get(device_id):
    """Retrieve a device or raise if it does not exist."""
    return IMPL.device_get(device_id)


def connection_get_all(execution_id):
    """Get all connections for an execution_id"""
    return IMPL.connection_get_all(execution_id)


def device_create(values):
    """Create an autodiscovery device."""
    return IMPL.device_create(values)


def device_update(device_id, values):
    """Update a device or raise if it does not exist."""
    return IMPL.device_update(device_id, values)


def connection_create(values):
    """Create an autodiscovery connection."""
    return IMPL.connection_create(values)


def exclusion_list_create(values):
    """Create an exclusion list."""
    return IMPL.exclusion_list_create(values)


def exclusion_list_get(exclusion_list_id):
    """Retrieve an exclusion list or raise if it does not exist."""
    return IMPL.exclusion_list_get(exclusion_list_id)


def exclusion_list_get_all():
    """Retrieve all registered exclusion lists."""
    return IMPL.exclusion_list_get_all()


def exclusion_list_get_default():
    """
    Retrieve or create the default exclusion list.

    This provides a single default exclusion list that can be used until
    support for multiple exclusion lists is implemented in the frontend.
    """
    default_id = uuid.UUID(int=0)
    try:
        return exclusion_list_get(default_id)
    except DBEntityNotFoundError:
        return exclusion_list_create(dict(
            id=default_id,
            name="Default",
            description="The default exclusion list",
            items=[]
        ))


def exclusion_list_update(exclusion_list_id, values):
    """Update an exclusion list or raise if it does not exist."""
    return IMPL.exclusion_list_update(exclusion_list_id, values)


def exclusion_list_delete(exclusion_list_id):
    """Delete an exclusion list or raise if it does not exist."""
    return IMPL.exclusion_list_delete(exclusion_list_id)


def exclusion_list_item_create(values):
    """Create an exclusion list item."""
    return IMPL.exclusion_list_item_create(values)


def exclusion_list_item_get(exclusion_list_item_id):
    """Retrieve an exclusion list or raise if it does not exist."""
    return IMPL.exclusion_list_item_get(exclusion_list_item_id)


def exclusion_list_item_get_all(exclusion_list_id):
    """Retrieve all registered exclusion lists."""
    return IMPL.exclusion_list_item_get_all(exclusion_list_id)


def exclusion_list_item_update(exclusion_list_item_id, values):
    """Update an exclusion list or raise if it does not exist."""
    return IMPL.exclusion_list_item_update(exclusion_list_item_id, values)


def exclusion_list_item_delete(exclusion_list_item_id):
    """Delete an exclusion list item or raise if it does not exist."""
    return IMPL.exclusion_list_item_delete(exclusion_list_item_id)


def exclusion_list_item_delete_all(exclusion_list_id):
    """Delete all items on an exclusion list."""
    return IMPL.exclusion_list_item_delete_all(exclusion_list_id)
