"""
Implementation of SQLAlchemy backend.
"""

import sys
import threading
import functools

from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy.exc_filters import exception as sqlalchemy_exceptions
from oslo_log import log as logging

from magellan.common import exceptions as exception
from magellan.common import consts
from magellan.db.sqlalchemy import migration
from magellan.db.sqlalchemy import models

LOG = logging.getLogger(__name__)

_main_context_manager = None
_CONTEXT = threading.local()


def _get_main_context_manager():
    global _main_context_manager
    if not _main_context_manager:
        _main_context_manager = enginefacade.transaction_context()
    return _main_context_manager


def get_engine():
    return _get_main_context_manager().get_legacy_facade().get_engine()


def get_backend():
    """The backend is this module itself."""
    return sys.modules[__name__]


def read_session():
    return _get_main_context_manager().reader.using(_CONTEXT)


def write_session():
    return _get_main_context_manager().writer.using(_CONTEXT)


def db_sync(engine, version=None):
    """Migrate the database to `version` or the most recent version."""
    return migration.db_sync(engine, version=version)


def db_version(engine):
    """Display the current database version."""
    return migration.db_version(engine)


def translate_exceptions(func):
    """
    Decorator for translating SQL Alchemy exceptions into Magellan exceptions.
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except sqlalchemy_exceptions.DBDuplicateEntry:
            raise exception.DBDuplicateEntryError()

    return wrapped


@translate_exceptions
def job_get(job_id):
    """Get an autodiscovery job by id."""
    with read_session() as session:
        job = session.query(models.AutoDiscoveryJob).filter(
            models.AutoDiscoveryJob.id == job_id).one_or_none()
        if not job:
            raise exception.DBEntityNotFoundError()
        return job


@translate_exceptions
def job_get_all():
    """Get all available autodiscovery jobs."""
    with read_session() as session:
        return session.query(models.AutoDiscoveryJob).all()


@translate_exceptions
def job_create(values):
    """
    Create a new autodiscovery job.
    :param values: a job structure.
    :return: a created job object.
    """
    with write_session() as session:
        job = models.AutoDiscoveryJob()
        job.update(values.copy())
        session.add(job)
        return job


@translate_exceptions
def job_update(job_id, values):
    """
    Update an existing autodiscovery job.
    :param job_id: id of a job to update.
    :param values: new job parameters.
    :return: an updated job object
    """
    with write_session() as session:
        job = job_get(job_id)
        job.update(values.copy())
        job.save(session)
        return job


@translate_exceptions
def job_delete(job_id):
    """
    Delete a job.
    :param job_id: the id of a job to delete.
    """
    with write_session() as session:
        job = job_get(job_id)
        session.delete(job)


@translate_exceptions
def execution_get(execution_id):
    """
    Get an autodiscovery execution by id.
    :param execution_id: The id of the execution to get
    :return: Instance of the execution
    """
    with read_session() as session:
        execution = session.query(models.AutoDiscoveryExecution).filter(
            models.AutoDiscoveryExecution.id == execution_id
        ).one_or_none()
        if not execution:
            raise exception.DBEntityNotFoundError()
        return execution


@translate_exceptions
def execution_get_all():
    """
    Get all available autodiscovery executions.
    :return: List of all execution instances
    """
    with read_session() as session:
        return session.query(models.AutoDiscoveryExecution).all()


@translate_exceptions
def execution_get_all_by_job(job_id):
    """
    Get all available autodiscovery executions.
    :param job_id: The id of the job to retrieve executions for
    :return: List of all execution instances for a job
    """
    job_get(job_id)
    with read_session() as session:
        return session.query(models.AutoDiscoveryExecution).filter(
            models.AutoDiscoveryExecution.job_id == job_id
        ).order_by(models.AutoDiscoveryExecution.created_at.desc()).all()


@translate_exceptions
def execution_get_all_scheduled(username):
    """
    Retrieve all latest scheduled executions.
    :param username: The username whose scheduled executions to return
    :return: List of all executions with a 'scheduled' schedule_status
    """
    with read_session() as session:
        schedule_status = models.AutoDiscoveryExecution.schedule_status
        scheduled = consts.ScheduleStatus.scheduled
        status = models.AutoDiscoveryExecution.status
        completed = consts.ExecutionStatus.completed
        query = session.query(
            models.AutoDiscoveryExecution
        ).filter(
            schedule_status == scheduled
        ).filter(
            status == completed
        )
        if username:
            query = query.filter(
                models.AutoDiscoveryExecution.username == username
            )
        return query.all()


@translate_exceptions
def execution_create(values):
    """
    Create a new autodiscovery execution.
    :param values: a execution structure.
    :return: a created execution object.
    """
    try:
        with write_session() as session:
            execution = models.AutoDiscoveryExecution()
            execution.update(values.copy())
            session.add(execution)
            return execution
    except sqlalchemy_exceptions.DBReferenceError:
        raise exception.DBEntityNotFoundError()


@translate_exceptions
def execution_update(execution_id, values):
    """
    Update an existing autodiscovery execution.
    :param execution_id: id of a execution to update.
    :return: a canceled execution object
    """
    with write_session() as session:
        execution = session.query(models.AutoDiscoveryExecution).filter(
            models.AutoDiscoveryExecution.id == execution_id
        ).one_or_none()
        if not execution:
            raise exception.DBEntityNotFoundError()

        execution.update(values)
        execution.save(session)

        return execution


@translate_exceptions
def execution_cancel(execution_id):
    """
    Update an existing autodiscovery execution.
    :param execution_id: id of a execution to update.
    :return: a canceled execution object
    """
    with write_session() as session:
        execution = execution_get(execution_id)
        execution.update({"status": "canceled"})
        execution.save(session)
        return execution


@translate_exceptions
def execution_delete(execution_id):
    """
    Delete an execution.
    :param execution_id: the id of a execution to delete.
    """
    with write_session() as session:
        execution = execution_get(execution_id)
        session.delete(execution)


@translate_exceptions
def device_get_all(execution_id):
    """
    Get all available autodiscovery devices for an execution.
    :return: List of all devices for an execution
    """
    with read_session() as session:

        execution = session.query(models.AutoDiscoveryExecution).filter(
            models.AutoDiscoveryExecution.id == execution_id
        ).one_or_none()

        if execution is None:
            raise exception.DBEntityNotFoundError()

        return execution.devices


@translate_exceptions
def device_get(device_id):
    """
    Get an discovered device by id.
    :param device_id: The id of the device to get
    :return: Instance of the device
    """
    with read_session() as session:
        device = session.query(models.AutoDiscoveryDevice).filter(
            models.AutoDiscoveryDevice.id == device_id
        ).one_or_none()
        if not device:
            raise exception.DBEntityNotFoundError()
        return device


@translate_exceptions
def connection_get_all(execution_id):
    """
    Get all available autodiscovery connections for an execution.
    :return: List of all connections for an execution
    """
    with read_session() as session:

        execution = session.query(models.AutoDiscoveryExecution).filter(
            models.AutoDiscoveryExecution.id == execution_id
        ).one_or_none()

        if execution is None:
            raise exception.DBEntityNotFoundError()

        return execution.connections


@translate_exceptions
def device_create(values):
    """
    Create a new autodiscovery device.
    :param values: a device structure.
    :return: a created device object.
    """
    with write_session() as session:
        device = models.AutoDiscoveryDevice()
        device.update(values.copy())
        session.add(device)
        return device


@translate_exceptions
def device_update(device_id, values):
    """
    Update an existing autodiscovery device.
    :param device_id: id of a device to update.
    :param values: updated device values.
    :return: an updated device object.
    """
    with write_session() as session:
        device = session.query(models.AutoDiscoveryDevice).filter(
            models.AutoDiscoveryDevice.id == device_id
        ).one_or_none()
        if not device:
            raise exception.DBEntityNotFoundError()

        device.update(values)
        device.save(session)

        return device


@translate_exceptions
def connection_create(values):
    """
    Create a new autodiscovery connection.
    :param values: a connection structure.
    :return: a created connection object.
    """
    with write_session() as session:
        connection = models.AutoDiscoveryConnection()
        connection.update(values.copy())
        session.add(connection)
        return connection


@translate_exceptions
def exclusion_list_get(exclusion_list_id):
    """Get an exclusion list by ID."""
    with read_session() as session:
        exclusion_list = session.query(
            models.AutoDiscoveryExclusionList).filter(
                models.AutoDiscoveryExclusionList.id ==
                exclusion_list_id).one_or_none()
        if not exclusion_list:
            raise exception.DBEntityNotFoundError()
        return exclusion_list


@translate_exceptions
def exclusion_list_get_all():
    """Get all available exclusion lists."""
    with read_session() as session:
        return session.query(models.AutoDiscoveryExclusionList).all()


@translate_exceptions
def exclusion_list_create(values):
    """
    Create a new exclusion list.
    :param values: a exclusion list structure.
    :return: a created exclusion list object.
    """
    with write_session() as session:
        exclusion_list = models.AutoDiscoveryExclusionList()
        items = values.pop("items")
        exclusion_list.update(values)
        session.add(exclusion_list)
        exclusion_list.update(values)
        exclusion_list.items = []  # Init to avoid lazy load later
        for address in items:
            exclusion_list_item = exclusion_list_item_create(address)
            exclusion_list.items.append(exclusion_list_item)
        return exclusion_list


@translate_exceptions
def exclusion_list_update(exclusion_list_id, values):
    """
    Update an existing autodiscovery exclusion list.
    :param exclusion_list_id: id of the exclusion list to update.
    :param values: new exclusion list parameters.
    :return: an updated exclusion list object
    """

    items = values.pop("items")
    with write_session() as session:
        exclusion_list_item_delete_all(exclusion_list_id)
        exclusion_list = exclusion_list_get(exclusion_list_id)
        for address in items:
            exclusion_list_item = exclusion_list_item_create(address)
            exclusion_list.items.append(exclusion_list_item)
        exclusion_list.update(values)
        exclusion_list.save(session)
        return exclusion_list


@translate_exceptions
def exclusion_list_delete(exclusion_list_id):
    """
    Delete an exclusion list.
    :param exclusion_list_id: id of the exclusion list to delete.
    """
    with write_session() as session:
        exclusion_list = exclusion_list_get(exclusion_list_id)
        session.delete(exclusion_list)


@translate_exceptions
def exclusion_list_item_create(values):
    """
    Create a new exclusion list item.
    :param values: a exclusion list item structure.
    :return: a created exclusion list object.
    """
    with write_session() as session:
        exclusion_list_item = models.AutoDiscoveryExclusionListItem()
        exclusion_list_item.update(values.copy())
        session.add(exclusion_list_item)
        return exclusion_list_item


@translate_exceptions
def exclusion_list_item_get(exclusion_list_item_id):
    """Get an exclusion list item by id."""
    with read_session() as session:
        exclusion_list = session.query(
            models.AutoDiscoveryExclusionListItem).filter(
                models.AutoDiscoveryExclusionListItem.id ==
                exclusion_list_item_id).one_or_none()
        if not exclusion_list:
            raise exception.DBEntityNotFoundError()
        return exclusion_list


@translate_exceptions
def exclusion_list_item_get_all(exclusion_list_id):
    """Get all exclusion list items by exclusion list id."""
    with read_session() as session:
        return session.query(
            models.AutoDiscoveryExclusionListItem).filter(
                models.AutoDiscoveryExclusionList.id ==
                exclusion_list_id).all()


@translate_exceptions
def exclusion_list_item_update(exclusion_list_item_id, values):
    """
    Update an exclusion list item.
    :param exclusion_list_id: id of the exclusion list item to update.
    :param values: new exclusion list item parameters.
    :return: an updated exclusion list item object
    """
    with write_session() as session:
        exclusion_list_item = exclusion_list_item_get(exclusion_list_item_id)
        exclusion_list_item.update(values.copy())
        exclusion_list_item.save(session)
        return exclusion_list_item


@translate_exceptions
def exclusion_list_item_delete(exclusion_list_item_id):
    """
    Delete an exclusion list item.
    :param exclusion_list_item_id: id of the exclusion list item to delete.
    """
    with write_session() as session:
        exclusion_list_item = exclusion_list_item_get(exclusion_list_item_id)
        session.delete(exclusion_list_item)


@translate_exceptions
def exclusion_list_item_delete_all(exclusion_list_id):
    """
    Delete all items on a exclusion list.
    :param exclusion_list_id: id of the exclusion list whose items to delete.
    """
    with write_session() as session:
        session.query(
            models.AutoDiscoveryExclusionListItem).filter(
                models.AutoDiscoveryExclusionList.id ==
                exclusion_list_id).delete(synchronize_session=False)
