import datetime
import ipaddress
import os
import re

import eventlet
import greenlet

from ipaddress import ip_address
from oslo_config import cfg
from oslo_db.exception import DBReferenceError
from oslo_log import log as logging

from magellan import db
from magellan.common.consts import (
    ScanMethod,
    ExecutionStatus,
    ScheduleStatus,
    DeviceStatus,
    SnmpAuthProtocol,
    SnmpPrivProtocol,
)
from magellan.common.exceptions import (
    DBEntityNotFoundError,
    MagellanError,
    LMDException,
)
from magellan.db.datafiles import get_ip_list_file
from magellan.engine import lmd
from magellan.engine.threadmanager import threadmanager
from pyautodiscover import (
    link_discover_seed, info, device_discover_range, link_discover_single,
    device_discover_list
)
from pyautodiscover.backends.snmp import (
    SNMPBackend, PrivInfo, PrivProto, AuthInfo, AuthProto, SNMPInfo
)
from pyautodiscover.backends.fping import FpingBackend
from pyautodiscover.backends.dns import DNSBackend
from pyautodiscover.consts import ScanMethod as pyAutoScanMethod
from pyautodiscover.utils.IPAddress import (
    is_address_within_range, is_fqdn
)

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
RE_WORD = re.compile("[^\\s]+")


def start_discovery(execution, job, exclusion_list=None):
    """ Start an execution and return the created green thread. """
    gt = threadmanager.start(_run_discovery, execution, job, exclusion_list)
    gt.execution_id = execution.id
    return gt


def stop_discovery(execution):
    """
    Stop a given execution from running and return its green thread, if stopped
    else None.
    """
    for gt in threadmanager.group.threads[:]:
        try:
            thread_id = gt.execution_id
        except AttributeError:
            continue
        if thread_id == execution.id:
            gt.stop()
            return gt
    return None


def get_snmp_info(snmp_options):
    if snmp_options.get('version') == 'v1':
        snmpinfo = SNMPInfo.SnmpV1(snmp_options.get('community'))
    elif snmp_options.get('version') == 'v2c':
        snmpinfo = SNMPInfo.SnmpV2c(snmp_options.get('community'))
    elif snmp_options.get('version') == 'v3':
        if snmp_options.get('authpass') is not None:
            ai = AuthInfo(
                AuthProto.by_name(snmp_options.get(
                    'authprotocol',
                    SnmpAuthProtocol.SHA.value,
                )),
                snmp_options.get('authpass')
            )
            if snmp_options.get('privpass') is not None:
                pi = PrivInfo(
                    PrivProto.by_name(snmp_options.get(
                        'privprotocol',
                        SnmpPrivProtocol.AES.value,
                    )),
                    snmp_options.get('privpass')
                )
                snmpinfo = SNMPInfo.SnmpV3AuthPriv(
                    snmp_options.get('secname'), ai, pi)
            else:
                snmpinfo = SNMPInfo.SnmpV3AuthNoPriv(
                    snmp_options.get('secname'), ai)
        else:
            snmpinfo = SNMPInfo.SnmpV3NoAuthNoPriv(
                snmp_options.get('secname'))
    return snmpinfo


def get_snmp_backend(snmp_options, exclusion_list=None):
    data_dir = CONF.get('data_dir')
    mib_dir = os.path.join(data_dir, 'mibs') if data_dir else None

    return SNMPBackend(
        get_snmp_info(snmp_options),
        mib_dir=mib_dir,
        use_arp=not CONF.discovery.no_arp,
        timeout=CONF.discovery.snmp_timeout,
        retries=CONF.discovery.snmp_retries,
        exclusion_list=exclusion_list,
    )


def create_or_update_device(backend, address, execution_id, devices):
    if address not in devices:
        devices[address] = db.api.device_create({
            'execution_id': execution_id,
            'address': str(address),
            'details': info(backend, address)
        })
    elif devices[address].details is None:
        devices[address] = db.api.device_update(
            devices[address].id, {
                'details': info(backend, address)
            }
        )


def add_links_in_db(backend, link, execution_id, devices, ip_filter=None):

    if link.address_a and (ip_filter is None or ip_filter(link.address_a)):
        create_or_update_device(backend, link.address_a, execution_id, devices)

    if link.address_b and (ip_filter is None or ip_filter(link.address_b)):
        create_or_update_device(backend, link.address_b, execution_id, devices)

    if link.address_a in devices and link.address_b in devices:
        db.api.connection_create({
            'execution_id': execution_id,
            'device1_id': devices[link.address_a].id,
            'device2_id': devices[link.address_b].id
        })


def get_hostname_address_mapping(ip_list):
    address_to_hostname = dict()
    for line in ip_list.splitlines():
        if '#' in line:
            line = line[:line.index('#')]
        for ip in line.split(' '):
            if ip and is_fqdn(ip):
                address_to_hostname[ip_address(
                    DNSBackend().get_host_address_by_name(ip))] = ip
    return address_to_hostname


def _run_discovery(execution, job, exclusion_list=None):
    """ Starts an autodiscovery job based on an execution """

    timeout = eventlet.Timeout(job.timeout)
    status = None
    devices = dict()
    address_to_hostname = dict()

    try:
        job = db.api.job_get(execution.job_id)
        if job.method == ScanMethod.seed_router:
            backend = get_snmp_backend(job.options['snmp'],
                                       exclusion_list=exclusion_list)
            links = link_discover_seed(backend, ipaddress.ip_address(
                job.options.get('seed_router')))

            for link in links:
                add_links_in_db(backend, link, execution.id, devices)

        elif job.method in (ScanMethod.ip_range, ScanMethod.ip_list):

            icmp_devices = None
            ip_filter = None

            if job.method == ScanMethod.ip_list:
                icmp_backend = FpingBackend(
                    pyAutoScanMethod.ip_list,
                    exclusion_list,
                    timeout=job.timeout,
                )
                address_to_hostname = get_hostname_address_mapping(
                    job.options['ip_list']
                )
                ip_list_file = get_ip_list_file(job.id)
                if os.path.isfile(ip_list_file):
                    icmp_devices = device_discover_list(
                        icmp_backend,
                        ip_list_file
                    )
                else:
                    raise MagellanError(
                        "Job '{}': ip list file doesn't exist: {}".format(
                            job.name, ip_list_file))

                def ip_filter(ip):
                    ip = str(ip)
                    for match in re.finditer(RE_WORD, job.options['ip_list']):
                        match = match.group()
                        if ip_address(ip) in address_to_hostname:
                            return True
                        elif any(c in match for c in ',-/'):
                            if is_address_within_range(ip, match):
                                return True
                        elif ip == match:
                            return True
                    return False
            else:
                ip_range = job.options['ip_range']
                icmp_backend = FpingBackend(
                    pyAutoScanMethod.ip_range,
                    exclusion_list,
                    timeout=job.timeout,
                )
                icmp_devices = device_discover_range(icmp_backend, ip_range)

                def ip_filter(ip):
                    return is_address_within_range(str(ip), ip_range)

            # Add devices from icmp scan
            for device in icmp_devices:
                if device not in devices:
                    devices[device] = db.api.device_create({
                        'execution_id': execution.id,
                        'address': str(device),
                        'details': {}
                    })
                    # Set details to None to possibly trigger SNMP lookup later
                    devices[device].details = None

            if 'snmp' in job.options:
                snmp_backend = get_snmp_backend(job.options['snmp'],
                                                exclusion_list=exclusion_list)
                for device in icmp_devices:
                    for link in link_discover_single(snmp_backend, device):
                        add_links_in_db(snmp_backend, link, execution.id,
                                        devices, ip_filter=ip_filter)
                    # in case no link is found, we still try to fetch snmp info
                    create_or_update_device(snmp_backend, device,
                                            execution.id, devices)

        if CONF.discovery.no_dns is False:
            # Do a reverse DNS lookup for devices that have no sysName yet
            dns_backend = DNSBackend()
            for address, device in devices.items():

                details = device.details or {}
                if "sysName" not in details:
                    dns_details = None
                    if address in address_to_hostname:
                        dns_details = {"sysName": address_to_hostname[address]}
                    else:
                        dns_details = info(dns_backend, address)
                    if dns_details:
                        details.update(dns_details)
                        devices[address] = db.api.device_update(
                                device.id, {"details": details}
                        )

        if execution.schedule_status == ScheduleStatus.scheduled:
            _delete_previous_scheduled_executions(execution)

        status = ExecutionStatus.completed
        LOG.info('The execution with id %s completed.', execution.id)

        if CONF.discovery.no_lmd is False:
            _update_device_status(execution)
    except (DBReferenceError, DBEntityNotFoundError):
        LOG.info(
            'The execution with id %s was deleted while running.', execution.id
        )
    except greenlet.GreenletExit:
        LOG.info('The execution with id %s was canceled.', execution.id)
        status = ExecutionStatus.canceled
    except eventlet.Timeout:
        LOG.info('The execution with id %s timed out.', execution.id)
        status = ExecutionStatus.interrupted
    except LMDException:
        LOG.warning(
            "The execution with id %s failed to query LMD for device status",
            execution.id
        )
        # Status already set to completed above.
    except Exception:
        LOG.error(
            'The execution with id %s raised an error', execution.id,
            exc_info=True
        )
        status = ExecutionStatus.error
    finally:
        timeout.cancel()
    if status:
        db.api.execution_update(execution.id, {
            'status': status,
            'complete_time': datetime.datetime.utcnow()
        })


def _delete_previous_scheduled_executions(execution):
    """ Delete previous scheduled executions of the same job. """
    for previous_execution in db.api.execution_get_all_by_job(
        execution.job_id
    ):
        if previous_execution.id == execution.id:
            continue
        if previous_execution.schedule_status == ScheduleStatus.unscheduled:
            continue
        db.api.execution_delete(previous_execution.id)


def _update_device_status(execution):
    devices = db.api.device_get_all(execution.id)

    lmd_query = "GET hosts\n"
    lmd_query += "Columns: address\n"
    for device in devices:
        lmd_query += f"Filter: address = {device.address}\n"
    lmd_query += f"OR: {len(devices)}\n"

    with lmd.LmdConnection() as conn:
        monitored = {address[0] for address in conn.json(lmd_query)}

    for device in devices:
        if device.address in monitored:
            status = DeviceStatus.monitored
        else:
            status = DeviceStatus.unmonitored
        db.api.device_update(device.id, {
            "status": status
        })
