import json
import socket

from oslo_config import cfg
from magellan.common.exceptions import LMDException


CONF = cfg.CONF


class LmdConnection:
    """
    Class for communicating with LMD over a unix socket.
    """

    def __init__(self, sockpath=None, timeout=None, encoding="utf-8"):
        """
        Initialize a new instance.

        :param sockpath: string path to the unix socket
        :param timeout: float number of seconds before a query timeouts
        :param encoding: string encoding to use
        """
        self.sockpath = sockpath or CONF.discovery.lmd_socket
        self.timeout = (
            CONF.discovery.lmd_timeout
            if timeout is None
            else timeout
        )
        self.encoding = encoding
        self._sock = None

    def connect(self):
        """
        Connect to the socket.
        """
        try:
            self.close()
            self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._sock.connect(self.sockpath)
            self._sock.settimeout(self.timeout)
            return self._sock
        except OSError as exc:
            raise LMDException(
                "Failed connecting to LMD: {}".format(exc)
            ) from exc

    def close(self):
        """
        Close the socket.
        """
        if self._sock:
            self._sock.close()
            self._sock = None

    def query(self, query):
        """
        Query LMD and return the result.

        :return: str with the result
        """
        if isinstance(query, str):
            query = query.encode(self.encoding)
        try:
            self._sock.sendall(query)
            self._sock.shutdown(socket.SHUT_WR)
            return self._read()
        except OSError as exc:
            raise LMDException("Failed to query LMD: {}".format(exc)) from exc

    def json(self, query):
        """
        Query LMD and return a list with the result.

        :return: list with the result
        """
        query += "OutputFormat: json\n"
        result = self.query(query)
        try:
            return json.loads(result) if result else []
        except ValueError as exc:
            raise LMDException(
                "Invalid JSON response from LMD: {}".format(exc)
            ) from exc

    def _read(self):
        buffer = bytearray()
        while True:
            data = self._sock.recv(1 << 10)
            if not data:
                break
            buffer.extend(data)
        return buffer.decode(self.encoding)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.close()
