import croniter

from datetime import datetime
import dateutil

from oslo_config import cfg
from oslo_log import log

from magellan.db import api as db_api
from magellan.common import exceptions, consts
from magellan.engine.threadmanager import threadmanager


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class JobScheduler:
    """
    Class for scheduling jobs.
    """

    def __init__(self, *args, **kwargs):
        self.tz = dateutil.tz.tzutc()
        self.timer = None
        self.last_tick = None

    def start(self):
        """
        Start the scheduler heartbeat.
        """
        if self.timer:
            raise RuntimeError("Scheduler already started")
        self.last_tick = datetime.now(tz=self.tz)
        initial_delay = self.get_delay(
            self.last_tick, CONF.engine.schedule_interval
        )
        self.timer = threadmanager.add_timer(
            CONF.engine.schedule_interval,
            JobScheduler.tick,
            self,
            initial_delay=initial_delay,
        )

    def tick(self):
        """
        Heartbeat method of the scheduler.

        This method will execute all scheduled jobs due since the last tick.
        """
        LOG.info("Checking for scheduled jobs...")
        this_tick = datetime.now(tz=self.tz)
        jobs = db_api.job_get_all()
        for job in jobs:
            try:
                if job.schedule:
                    cron = croniter.croniter(job.schedule, self.last_tick)
                    scheduled_at = datetime.fromtimestamp(
                        cron.get_next(), tz=self.tz
                    )
                    if scheduled_at <= this_tick:
                        self.run_scheduled_job(job)
            except croniter.CroniterBadCronError:
                LOG.error("Bad cron format for job %s", job.id)
            except Exception:
                LOG.error("Error while scheduling job", exc_info=True)
        self.last_tick = this_tick

    def run_scheduled_job(self, job):
        """
        Start a scheduled execution for a job.

        :param job: AutoDiscoveryJob to execute
        """
        LOG.info("Start scheduled execution of job [%s]", job.id)
        try:
            db_api.execution_create({
                'job_id': job.id,
                'name': job.name,
                'status': consts.ExecutionStatus.running,
                'schedule_status': consts.ScheduleStatus.scheduled,
                'description': '',
                'username': job.username,
            })
        except (exceptions.JobAlreadyRunningError, exceptions.DBError):
            LOG.error("Scheduled execution of job [%s] failed", job.id)

    @staticmethod
    def get_delay(last_tick, target_interval):
        """
        Calculate the delay required to get to the next even target_interval.
        E.g. if the target_interval is 1h, the next tick should happen at the
        next full hour.

        :param last_tick: datetime from where to calculate the delay
        :param target_interval: seconds of the regular interval
        """
        if target_interval > (60 * 60):
            return target_interval

        return (
            target_interval
            - (last_tick.minute * 60) % target_interval
            - last_tick.second
        )
