import oslo_messaging
import datetime

from oslo_config import cfg
from oslo_log import log as logging
from magellan.common import consts
from magellan.common import messaging as rpc_messaging
from magellan.common import exceptions
from magellan.db import api as db_api
from magellan.common.consts import ExecutionStatus


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class RpcServer(object):
    """ RPC server.
    This class gets registered as an RPC endpoint for engine specific calls.
    """

    def __init__(self, engine):

        self.engine = engine

        self.host = CONF.engine.host
        self.rpc_api_version = consts.RPC_API_VERSION
        self.topic = "%s_%s" % (CONF.messaging.exchange, "rpc")
        # The following are initialized here, but assigned in start() which
        # happens after the fork when spawning multiple worker processes
        self.target = None
        self._rpc_server = None

    def start(self):
        self.target = oslo_messaging.Target(
            version=self.rpc_api_version,
            server=self.host,
            topic=self.topic
        )
        self._rpc_server = rpc_messaging.get_rpc_server(self.target, self)
        self._rpc_server.start()

        self.cleanup_executions()
        LOG.info('Magellan RPC server started')

    def stop(self):
        if self._rpc_server:
            try:
                self._rpc_server.stop()
                self._rpc_server.wait()
                LOG.info('Magellan RPC server stopped')
            except Exception as ex:
                LOG.error('Failed to stop Magellan RPC server: %s', str(ex))

    # Job methods
    @staticmethod
    def cleanup_executions():
        executions = db_api.execution_get_all()
        for execution in executions:
            if execution.status == ExecutionStatus.running:
                db_api.execution_update(execution.id, {
                    'status': ExecutionStatus.error,
                    'complete_time': datetime.datetime.utcnow()
                })

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def job_create(self, ctxt, values):
        LOG.debug("Create job [%s]", values)
        return db_api.job_create(values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def job_get_all(self, ctxt):
        LOG.debug("Fetch all jobs.")
        return db_api.job_get_all()

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def job_get(self, ctxt, job_id):
        LOG.debug("Fetch job [id=%s]", job_id)
        return db_api.job_get(job_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def job_update(self, ctxt, job_id, values):
        LOG.debug("Update job [id=%s]", job_id)
        return db_api.job_update(job_id, values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def job_delete(self, ctxt, job_id):
        LOG.debug("Delete job [id=%s]", job_id)
        db_api.job_delete(job_id)

    # Execution methods

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_create(self, ctxt, values):
        LOG.debug("Create execution [%s]", values)
        return db_api.execution_create(values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_update(self, ctxt, execution_id, values):
        LOG.debug("Update execution [id=%s]", execution_id)
        return db_api.execution_update(execution_id, values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_delete(self, ctxt, execution_id):
        LOG.debug("Delete execution [id=%s]", execution_id)
        return db_api.execution_delete(execution_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_cancel(self, ctxt, execution_id):
        LOG.debug("Cancel execution [id=%s]", execution_id)
        return db_api.execution_cancel(execution_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_get(self, ctxt, execution_id):
        LOG.debug("Fetch execution [id=%s]", execution_id)
        return db_api.execution_get(execution_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_get_all(self, ctxt):
        LOG.debug("Fetch all executions")
        return db_api.execution_get_all()

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_get_all_by_job(self, ctxt, job_id):
        LOG.debug("Fetch all executions for job")
        return db_api.execution_get_all_by_job(job_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def execution_get_all_scheduled(self, ctxt, username):
        LOG.debug("Fetch the latest scheduled executions [user=%s]", username)
        return db_api.execution_get_all_scheduled(username)

    # Device methods

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def device_get_all(self, ctxt, execution_id):
        LOG.debug("Fetch all devices")
        return db_api.device_get_all(execution_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def device_get(self, ctxt, device_id):
        LOG.debug("Fetch device [id=%s]", device_id)
        return db_api.device_get(device_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def device_update(self, ctxt, device_id, values):
        LOG.debug("Update device [id=%s]", device_id)
        return db_api.device_update(device_id, values)

    # Connection methods

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def connection_get_all(self, ctxt, execution_id):
        LOG.debug("Fetch all connections")
        return db_api.connection_get_all(execution_id)

    # Exclusion list methods

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_create(self, ctxt, values):
        LOG.debug("Create exclusion list [%s]", values)
        return db_api.exclusion_list_create(values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_get_all(self, ctxt):
        LOG.debug("Fetch all exclusion lists.")
        return db_api.exclusion_list_get_all()

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_get(self, ctxt, exclusion_list_id):
        LOG.debug("Fetch exclusion list [id=%s]", exclusion_list_id)
        return db_api.exclusion_list_get(exclusion_list_id)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_get_default(self, ctxt):
        LOG.debug("Fetch default exclusion list")
        return db_api.exclusion_list_get_default()

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_update(self, ctxt, exclusion_list_id, values):
        LOG.debug("Update exclusion list [id=%s]", exclusion_list_id)
        return db_api.exclusion_list_update(exclusion_list_id, values)

    @oslo_messaging.expected_exceptions(
        exceptions.MagellanException,
        exceptions.MagellanError
    )
    def exclusion_list_delete(self, ctxt, exclusion_list_id):
        LOG.debug("Delete exclusion list [id=%s]", exclusion_list_id)
        db_api.exclusion_list_delete(exclusion_list_id)
