
import unittest

from wsme.rest import json as wjson
from magellan.api.controllers.v1.resources import Job, SNMPv1, SNMPv2, \
                                                  SNMPv3AuthPriv, \
                                                  SNMPv3AuthNoPriv, \
                                                  SNMPv3NoAuthNoPriv
from magellan.common import consts


class TestJobResource(unittest.TestCase):

    def test_job_invalid_ip_address(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127,0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'secname': 'testuser'
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v1(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v1',
                    'community': 'public'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv1)

    def test_job_snmp_v1_has_default_community(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v1'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv1)
        assert job.options.snmp.community == 'public'

    def test_job_snmp_v2(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v2c',
                    'community': 'public'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv2)

    def test_job_snmp_v2_has_default_community(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v2c'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv2)
        assert job.options.snmp.community == 'public'

    def test_job_snmp_v3(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v3',
                    'secname': 'testuser',
                    'authpass': 'testauthpass',
                    'privpass': 'testprivpass'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv3AuthPriv)
        # Check defaults
        assert job.options.snmp.authprotocol == consts.SnmpAuthProtocol.SHA
        assert job.options.snmp.privprotocol == consts.SnmpPrivProtocol.AES

    def test_job_snmp_v3_explicit_protocols(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v3',
                    'secname': 'testuser',
                    'authpass': 'testauthpass',
                    'privpass': 'testprivpass',
                    'authprotocol': 'SHA',
                    'privprotocol': 'AES',
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv3AuthPriv)
        assert job.options.snmp.authprotocol == consts.SnmpAuthProtocol.SHA
        assert job.options.snmp.privprotocol == consts.SnmpPrivProtocol.AES

    def test_job_snmp_v3_alt_protocols(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v3',
                    'secname': 'testuser',
                    'authpass': 'testauthpass',
                    'privpass': 'testprivpass',
                    'authprotocol': 'MD5',
                    'privprotocol': '3DES',
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv3AuthPriv)
        assert job.options.snmp.authprotocol == consts.SnmpAuthProtocol.MD5
        assert job.options.snmp.privprotocol == consts.SnmpPrivProtocol.DES3

    def test_job_snmp_v3AuthNoPriv(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v3',
                    'secname': 'testuser',
                    'authpass': 'testprivpass'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv3AuthNoPriv)

    def test_job_snmp_v3NoAuthNoPriv(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'seed_router',
            'options': {
                'seed_router': '127.0.0.1',
                'snmp': {
                    'version': 'v3',
                    'secname': 'testuser'
                }
            },
            'timeout': 3600
        })
        assert isinstance(job.options.snmp, SNMPv3NoAuthNoPriv)

    def test_job_snmp_v3_missing_data(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'seciname': 'testuser'
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_type_for_secname(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'secname': 123
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_value_for_version(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v4',
                        'secname': 'testuser'
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_type_for_version(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 1,
                        'secname': 'testuser'
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_type_for_authpass(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'secname': 'testuser',
                        'authpass': 123
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_type_for_privpass(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'secname': 'testuser',
                        'authpass': '123',
                        'privpass': 123
                    }
                },
                'timeout': 3600
            })

    def test_job_snmp_v3_invalid_value_for_authprotocol(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'seed_router',
                'options': {
                    'seed_router': '127.0.0.1',
                    'snmp': {
                        'version': 'v3',
                        'secname': 'testuser',
                        'authpass': '123',
                        'privpass': '123',
                        'authprotocol': 'badproto',
                    }
                },
                'timeout': 3600
            })

    def test_job_ip_range(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'ip_range',
            'options': {
                'ip_range': '127.0.0.-3,4,5,250-'
            },
            'timeout': 3600
        })
        assert job.options.ip_range == '127.0.0.-3,4,5,250-'

    def test_job_subnet(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'ip_range',
            'options': {
                'ip_range': '127.0.0.0/24'
            },
            'timeout': 3600
        })
        assert job.options.ip_range == '127.0.0.0/24'

    def test_job_ip_list(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'ip_list',
            'options': {
                'ip_list': '127.0.0.1 127.0.0.2 127.0.0.3'
            },
            'timeout': 3600
        })
        assert job.options.ip_list == '127.0.0.1 127.0.0.2 127.0.0.3'

    def test_job_ip_list_with_hostname(self):
        job = wjson.fromjson(Job, {
            'id': '123e4567-e89b-12d3-a456-426655440000',
            'name': 'job',
            'description': 'This is a long and useless',
            'method': 'ip_list',
            'options': {
                'ip_list': '127.0.0.1 127.0.0.2 127.0.0.3 '
                           'www.op5.com'
            },
            'timeout': 3600
        })
        assert job.options.ip_list == '127.0.0.1 127.0.0.2 127.0.0.3 ' \
                                      'www.op5.com'

    def test_job_ip_list_invalid_list(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'ip_list',
                'options': {
                    'ip_list': '127.0.0.1 127.0.0. 127.0.0.3'
                },
                'timeout': 3600
            })

    def test_job_ip_list_invalid_list_with_hostname(self):
        with self.assertRaisesRegex(
                ValueError, "does not match any variant of Job Options"):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'ip_list',
                'options': {
                    'ip_list': '127.0.0.1 -www.op5.com 127.0.0.3 '
                },
                'timeout': 3600
            })

    def test_mismatching_method_options(self):
        with self.assertRaisesRegex(
            ValueError, "Mismatching job method and options"
        ):
            wjson.fromjson(Job, {
                'id': '123e4567-e89b-12d3-a456-426655440000',
                'name': 'job',
                'description': 'This is a long and useless',
                'method': 'ip_range',
                'options': {
                    'ip_list': '127.0.0.1 127.0.0.2 127.0.0.3'
                },
                'timeout': 3600
            })
