import datetime
import mock

import eventlet

from magellan.common import consts, exceptions
from magellan.db import api as db_api
from magellan.engine import discover, lmd
from magellan.tests import base as test_base
from pyautodiscover import Link
from pyautodiscover.backends.dns import DNSBackend
from ipaddress import ip_address

PUBLIC_V2C_SEED_ROUTER_JOB = {
    'name': 'my_job1',
    'description': 'empty',
    'method': consts.SCAN_METHOD_SEED_ROUTER,
    'options': {
        'seed_router': '192.168.1.1',
        'snmp': {
            'version': 'v2c',
            'community': 'public'
        }
    },
    'timeout': 1800,
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 0, 0)
}

PRIVATE_V1_SEED_ROUTER_JOB = {
    'name': 'my_job2',
    'description': 'my description',
    'method': consts.SCAN_METHOD_SEED_ROUTER,
    'options': {
        'seed_router': '192.168.1.1',
        'snmp': {
            'version': 'v1',
            'community': 'private'
        }
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

IP_RANGE_JOB = {
    'name': 'my_job3',
    'description': 'my description',
    'method': consts.SCAN_METHOD_IP_RANGE,
    'options': {
        'ip_range': '192.168.1.1-10'
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

SUBNET_JOB = {
    'name': 'my_job4',
    'description': 'my description',
    'method': consts.SCAN_METHOD_IP_RANGE,
    'options': {
        'ip_range': '192.168.1.0/30'
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

PUBLIC_V2C_IP_RANGE_JOB = {
    'name': 'my_job5',
    'description': 'my description',
    'method': consts.SCAN_METHOD_IP_RANGE,
    'options': {
        'ip_range': '192.168.1.1-5',
        'snmp': {
            'version': 'v2c',
            'community': 'public'
        }
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

PUBLIC_V2C_IP_LIST_JOB = {
    'name': 'my_job6',
    'description': 'my description',
    'method': consts.SCAN_METHOD_IP_LIST,
    'options': {
        'ip_list': '192.168.1.1 192.168.1.2 192.168.1.3 www.op5.com',
        'snmp': {
            'version': 'v2c',
            'community': 'public'
        }
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

V3_IP_LIST_JOB = {
    'name': 'my_job7',
    'description': 'my description',
    'method': consts.SCAN_METHOD_SEED_ROUTER,
    'options': {
        'seed_router': '192.168.1.1',
        'snmp': {
            'version': 'v3',
            'secname': 'kaka',
            'authpass': 'abc123',
            'privpass': 'abc123'
        }
    },
    'timeout': 3600,
    'schedule': '0 */4 * * *',
    'updated_at': None,
    'created_at': datetime.datetime(2017, 12, 1, 15, 1, 0)
}

DISCOVERY_EXECUTIONS = [
    {
        'job_id': None,
        'name': 'my_execution1',
        'description': 'my description',
        'status': consts.ExecutionStatus.running
    },
    {
        'job_id': None,
        'name': 'my_execution2',
        'description': 'my description',
        'status': consts.ExecutionStatus.running
    }
]


def add_job_id(data, job):
    copy = data.copy()
    copy['job_id'] = job['id']
    return copy


class DiscoveryExecutionTest(test_base.SqlAlchemyTest):
    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_execution_create(self, discover_func, info, backend_cls):
        self.override_config("no_lmd", True, group="discovery")

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4', '192.168.1.5']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[1].id),
                (devices[0].id, devices[2].id)
            ]
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_execution_create_v3(self, discover_func, info, backend_cls):
        self.override_config("no_lmd", True, group="discovery")

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(V3_IP_LIST_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4', '192.168.1.5']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[1].id),
                (devices[0].id, devices[2].id)
            ]
        )
        snmpinfo = backend_cls.mock_calls[0].args[0]
        self.assertEqual(
            snmpinfo.secname,
            V3_IP_LIST_JOB['options']['snmp']['secname']
        )
        self.assertEqual(snmpinfo.authinfo.proto.as_name(), 'SHA')  # Default
        self.assertEqual(snmpinfo.privinfo.proto.as_name(), 'AES')  # Default

    @mock.patch.object(discover, 'LOG')
    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_execution_delete(self, discover_func, info, backend_cls, logger):
        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )
        db_api.execution_delete(execution.id)

        # Give the discovery green thread a chance to run (and fail)
        eventlet.sleep(0.1)

        self.assertRaises(exceptions.DBEntityNotFoundError,
                          db_api.device_get_all,
                          execution['id'])
        self.assertRaises(exceptions.DBEntityNotFoundError,
                          db_api.connection_get_all,
                          execution['id'])
        self.assertEqual(logger.info.call_count, 1)

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_execution_cancel(self, discover_func, info, backend_cls):
        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            eventlet.sleep(0.2)
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        db_api.execution_cancel(execution['id'])

        execution = db_api.execution_get(execution['id'])
        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(execution['status'], consts.ExecutionStatus.canceled)
        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[1].id)
            ]
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_execution_timeout(self, discover_func, info, backend_cls):
        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            eventlet.sleep(0.2)
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}
        job_data = PUBLIC_V2C_SEED_ROUTER_JOB.copy()
        job_data['timeout'] = 0.1  # We cheat and use a float to test faster

        job = db_api.job_create(job_data)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.2)

        execution = db_api.execution_get(execution['id'])
        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(
            execution['status'],
            consts.ExecutionStatus.interrupted
        )
        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[1].id)
            ]
        )

    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_range')
    def test_range_execution_create(self, device_discover_range,
                                    info, backend_cls):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_range.return_value = [ip_address('192.168.1.3'),
                                              ip_address('192.168.1.4'),
                                              ip_address('192.168.1.5')]
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(IP_RANGE_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4', '192.168.1.5']
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_range')
    @mock.patch.object(discover, 'link_discover_single')
    def test_range_execution_create_filter(self, link_discover_func,
                                           device_discover_func,
                                           info, backend_cls, backend_snmp):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_func.return_value = [ip_address('192.168.1.3'),
                                             ip_address('192.168.1.5')]

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.6'))

        link_discover_func.return_value = generate_links()

        info.return_value = {'sysName': 'foobar'}
        self.assertIs(discover.device_discover_range, device_discover_func)
        self.assertIs(discover.link_discover_single, link_discover_func)

        job = db_api.job_create(PUBLIC_V2C_IP_RANGE_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        # self.assertEqual(logger.info.call_count, 1)
        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.5', '192.168.1.4']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[2].id),
                (devices[0].id, devices[1].id)
            ]
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_range')
    @mock.patch.object(discover, 'link_discover_single')
    def test_range_execution_icmp_first(self, link_discover_func,
                                        device_discover_func,
                                        info, backend_cls, backend_snmp):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_func.return_value = [ip_address('192.168.1.3'),
                                             ip_address('192.168.1.5')]

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            eventlet.sleep(0.1)
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.6'))

        link_discover_func.return_value = generate_links()

        info.return_value = {'sysName': 'foobar'}
        self.assertIs(discover.device_discover_range, device_discover_func)
        self.assertIs(discover.link_discover_single, link_discover_func)

        job = db_api.job_create(PUBLIC_V2C_IP_RANGE_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        # self.assertEqual(logger.info.call_count, 1)
        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        # all ICMP device should be found now in the DB
        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.5', '192.168.1.4']
        )
        # only one link should be discovered by SNMP
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[2].id)
            ]
        )

    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_range')
    def test_subnet_execution_create(self, device_discover_range,
                                     info, backend_cls):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_range.return_value = [ip_address('192.168.1.3'),
                                              ip_address('192.168.1.4'),
                                              ip_address('192.168.1.5')]
        info.return_value = {'sysName': 'foobar'}

        job = db_api.job_create(SUBNET_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.4', '192.168.1.5']
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_range')
    @mock.patch.object(discover, 'link_discover_single')
    def test_range_execution_snmp_info_even_no_link(self, link_discover_func,
                                                    device_discover_func,
                                                    info, backend_cls,
                                                    backend_snmp):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_func.return_value = [ip_address('192.168.1.3'),
                                             ip_address('192.168.1.5')]

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))

        link_discover_func.return_value = generate_links()

        info.return_value = {'sysName': 'foobar'}
        self.assertIs(discover.device_discover_range, device_discover_func)
        self.assertIs(discover.link_discover_single, link_discover_func)

        job = db_api.job_create(PUBLIC_V2C_IP_RANGE_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        # self.assertEqual(logger.info.call_count, 1)
        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        # all ICMP device should be found now in the DB
        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.3', '192.168.1.5', '192.168.1.4']
        )
        # only one link should be discovered by SNMP
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[0].id, devices[2].id)
            ]
        )
        # check device 192.168.1.5 also has snmp info even it has no link
        self.assertEqual(devices[2].details, info.return_value)

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'FpingBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'device_discover_list')
    @mock.patch.object(discover, 'link_discover_single')
    @mock.patch.object(discover, 'get_hostname_address_mapping')
    def test_ip_list_execution(self, get_hostname_address_mapping_func,
                               link_discover_func,
                               device_discover_func,
                               info, backend_cls, backend_snmp):
        self.override_config("no_lmd", True, group="discovery")

        device_discover_func.return_value = [ip_address('192.168.1.1'),
                                             ip_address('192.168.1.3')]

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.2'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.100'))

        link_discover_func.return_value = generate_links()

        get_hostname_address_mapping_func.return_value = \
            {ip_address('192.168.1.100'): 'www.op5.com'}

        info.return_value = {'sysName': 'foobar'}
        self.assertIs(discover.device_discover_list, device_discover_func)
        self.assertIs(discover.link_discover_single, link_discover_func)

        job = db_api.job_create(PUBLIC_V2C_IP_LIST_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])
        connections = db_api.connection_get_all(execution['id'])

        self.assertEqual(
            [device.address for device in devices],
            ['192.168.1.1', '192.168.1.3', '192.168.1.2', '192.168.1.100']
        )
        self.assertEqual(
            [
                (connection.device1_id, connection.device2_id)
                for connection in connections
            ],
            [
                (devices[1].id, devices[2].id),
                (devices[1].id, devices[3].id),
            ]
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_dns_info(self, discover_func, backend_cls):
        self.override_config("no_lmd", True, group="discovery")

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()

        def mocked_info(backend, address):
            if isinstance(backend, DNSBackend):
                return {'sysName': 'dns-hostname'}
            elif address == ip_address('192.168.1.3'):
                return {'sysName': 'snmp-hostname', 'sysDescr': 'snmp-descr'}
            return {}

        with mock.patch.object(discover, 'info', mocked_info):
            job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
            execution = db_api.execution_create(
                add_job_id(DISCOVERY_EXECUTIONS[0], job)
            )

            # Give the discovery green thread a chance to run
            eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])

        self.assertEqual(
            {
                '192.168.1.3': {
                    'sysName': 'snmp-hostname', 'sysDescr': 'snmp-descr'
                },
                '192.168.1.4': {'sysName': 'dns-hostname'},
                '192.168.1.5': {'sysName': 'dns-hostname'}
            },
            {device.address: device.details for device in devices}
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'link_discover_seed')
    def test_disabled_dns_info(self, discover_func, backend_cls):
        self.override_config("no_lmd", True, group="discovery")
        self.override_config("no_dns", True, group="discovery")

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))
        discover_func.return_value = generate_links()

        def mocked_info(backend, address):
            if isinstance(backend, DNSBackend):
                return {'sysName': 'dns-hostname'}
            elif address == ip_address('192.168.1.3'):
                return {'sysName': 'snmp-hostname', 'sysDescr': 'snmp-descr'}
            return {}

        with mock.patch.object(discover, 'info', mocked_info):
            job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
            execution = db_api.execution_create(
                add_job_id(DISCOVERY_EXECUTIONS[0], job)
            )

            # Give the discovery green thread a chance to run
            eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])

        self.assertEqual(
            {
                '192.168.1.3': {
                    'sysName': 'snmp-hostname', 'sysDescr': 'snmp-descr'
                },
                '192.168.1.4': {},
                '192.168.1.5': {}
            },
            {device.address: device.details for device in devices}
        )

    @mock.patch.object(discover, 'SNMPBackend')
    @mock.patch.object(discover, 'info')
    @mock.patch.object(discover, 'link_discover_seed')
    @mock.patch.object(lmd, 'socket')
    def test_update_device_status(
        self,
        lmd_socket,
        discover_func,
        info,
        backend_cls
    ):
        self.override_config("no_lmd", False, group="discovery")

        def generate_links():
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.4'))
            yield Link(ip_address('192.168.1.3'), ip_address('192.168.1.5'))

        discover_func.return_value = generate_links()
        info.return_value = {'sysName': 'foobar'}

        lmd_socket.socket = mock.Mock(return_value=mock.Mock(
            recv=mock.Mock(side_effect=[b'[["192.168.1.4"]]', b''])
        ))

        job = db_api.job_create(PUBLIC_V2C_SEED_ROUTER_JOB)
        execution = db_api.execution_create(
            add_job_id(DISCOVERY_EXECUTIONS[0], job)
        )

        # Give the discovery green thread a chance to run
        eventlet.sleep(0.1)

        devices = db_api.device_get_all(execution['id'])

        self.assertEqual(
            [
                (device.address, device.status) for device in devices
            ],
            [
                ('192.168.1.3', consts.DeviceStatus.unmonitored),
                ('192.168.1.4', consts.DeviceStatus.monitored),
                ('192.168.1.5', consts.DeviceStatus.unmonitored),
            ]
        )
