import datetime
import mock
import uuid

from magellan.engine import scheduler
from magellan.common.consts import ExecutionStatus, ScheduleStatus
from magellan.db.sqlalchemy.models import AutoDiscoveryJob


class MockDatetime(datetime.datetime):
    mock_now = None

    @classmethod
    def set_now(cls, now):
        cls.mock_now = now.astimezone()

    @classmethod
    def now(cls, *args, **kwargs):
        return cls.mock_now


@mock.patch.object(scheduler, 'db_api')
@mock.patch.object(scheduler, 'threadmanager')
@mock.patch.object(scheduler, 'datetime', new=MockDatetime)
def test_start(mock_tm, mock_db):
    now = datetime.datetime(2001, 9, 11, 8, 46)
    MockDatetime.set_now(now)
    job_scheduler = scheduler.JobScheduler()
    job_scheduler.start()
    assert job_scheduler.last_tick == now.astimezone()


@mock.patch.object(scheduler, 'db_api')
@mock.patch.object(scheduler, 'threadmanager')
@mock.patch.object(scheduler, 'datetime', new=MockDatetime)
def test_tick(mock_tm, mock_db):
    MockDatetime.set_now(datetime.datetime(2001, 9, 11, 8, 46))
    job_scheduler = scheduler.JobScheduler()
    job_scheduler.start()
    mock_db.job_get_all = mock.Mock(return_value=[
        AutoDiscoveryJob(
            id=uuid.UUID(int=1), name="Gurka", schedule="5 * * * *"
        ),
        AutoDiscoveryJob(
            id=uuid.UUID(int=2), name="Tomat", schedule="10 * * * *"
        ),
        AutoDiscoveryJob(
            id=uuid.UUID(int=3), name="Sallad", schedule="20 * * * *"
        ),
    ])
    MockDatetime.set_now(datetime.datetime(2001, 9, 11, 9, 1))
    job_scheduler.tick()
    mock_db.execution_create.assert_has_calls([])
    MockDatetime.set_now(datetime.datetime(2001, 9, 11, 9, 16))
    job_scheduler.tick()
    mock_db.execution_create.assert_has_calls([
        mock.call({
            'job_id': uuid.UUID(int=1),
            'name': 'Gurka',
            'status': ExecutionStatus.running,
            'description': '',
            'schedule_status': ScheduleStatus.scheduled,
            'username': None,
        }),
        mock.call({
            'job_id': uuid.UUID(int=2),
            'name': 'Tomat',
            'status': ExecutionStatus.running,
            'description': '',
            'schedule_status': ScheduleStatus.scheduled,
            'username': None,
        }),
    ])
    MockDatetime.set_now(datetime.datetime(2001, 9, 11, 9, 31))
    job_scheduler.tick()
    mock_db.execution_create.assert_has_calls([
        mock.call({
            'job_id': uuid.UUID(int=1),
            'name': 'Gurka',
            'status': ExecutionStatus.running,
            'description': '',
            'schedule_status': ScheduleStatus.scheduled,
            'username': None,
        }),
        mock.call({
            'job_id': uuid.UUID(int=2),
            'name': 'Tomat',
            'status': ExecutionStatus.running,
            'description': '',
            'schedule_status': ScheduleStatus.scheduled,
            'username': None,
        }),
        mock.call({
            'job_id': uuid.UUID(int=3),
            'name': 'Sallad',
            'status': ExecutionStatus.running,
            'description': '',
            'schedule_status': ScheduleStatus.scheduled,
            'username': None,
        }),
    ])


def test_get_delay():
    # Test that 1h interval return the next full hour
    tick = datetime.datetime(2001, 9, 11, 8, 46, 12)
    delay = scheduler.JobScheduler.get_delay(tick, 60 * 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 9, 0, 0
    )

    # Test shortest interval
    tick = datetime.datetime(2001, 9, 11, 8, 59, 59)
    delay = scheduler.JobScheduler.get_delay(tick, 60 * 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 9, 0, 0
    )

    # Test longest interval
    tick = datetime.datetime(2001, 9, 11, 9, 0, 0)
    delay = scheduler.JobScheduler.get_delay(tick, 60 * 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 10, 0, 0
    )

    # Test that 15 min interval return the next full quarter
    tick = datetime.datetime(2001, 9, 11, 8, 36, 12)
    delay = scheduler.JobScheduler.get_delay(tick, 15 * 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 8, 45, 0
    )

    # Test that 1 min interval return the next full minute
    tick = datetime.datetime(2001, 9, 11, 8, 36, 12)
    delay = scheduler.JobScheduler.get_delay(tick, 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 8, 37, 0
    )

    # Test that interval > 1h are ignored
    tick = datetime.datetime(2001, 9, 11, 8, 46, 12)
    delay = scheduler.JobScheduler.get_delay(tick, 60 * 60 + 15 * 60)
    assert tick + datetime.timedelta(0, delay) == datetime.datetime(
        2001, 9, 11, 10, 1, 12
    )
