#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# pylint: disable=E1102
# disable warning about the constructors in the types imported from
# backends.snmp not being callable

from pyautodiscover import (
    link_discover_seed,
    link_discover_single,
    device_discover_range,
    device_discover_list,
)
from pyautodiscover.backends.snmp import (
    SNMPBackend,
    SNMPInfo,
    AuthProto,
    PrivProto,
    AuthInfo,
    PrivInfo,
)
from ipaddress import ip_address
from pyautodiscover.backends.fping import FpingBackend
from pyautodiscover.consts import ScanMethod
import argparse
import csv
import sys
import json

import logging


logger = logging.getLogger()
handler = logging.StreamHandler()
formatter = logging.Formatter("%(name)-12s %(levelname)-8s %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.ERROR)


def ipaddress_parser(ip):
    try:
        return ip_address(ip)
    except ValueError:
        raise argparse.ArgumentTypeError("{} is not a valid IPv4 address".format(ip))


def exclusion_list_parser(exclusion_list):
    try:
        exclusion_list = json.loads(exclusion_list)
        if not isinstance(exclusion_list, list):
            raise ValueError()
        return exclusion_list
    except ValueError:
        raise argparse.ArgumentTypeError(
            "{} is not a valid JSON list".format(exclusion_list)
        )


def v1info(args):
    return SNMPInfo.SnmpV1(args.community)


def v2cinfo(args):
    return SNMPInfo.SnmpV2c(args.community)


def v3info(args):
    if args.auth_key is not None:
        ai = AuthInfo(
            AuthProto.by_name(args.auth_proto),
            args.auth_key
        )
        if args.priv_key is not None:
            pi = PrivInfo(
                PrivProto.by_name(args.priv_proto),
                args.priv_key
            )
            return SNMPInfo.SnmpV3AuthPriv(args.secname, ai, pi)
        return SNMPInfo.SnmpV3AuthNoPriv(args.secname, ai)
    return SNMPInfo.SnmpV3NoAuthNoPriv(args.secname)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""
Python discovery script.
Use SNMP to retrieve and recursively walk through neighbours of a
target address by LLDP information.
        """,
    )

    subparsers = parser.add_subparsers()

    parser_v1 = subparsers.add_parser("v1", help="Use SNMPv1")
    parser_v1.add_argument(
        "-c",
        "--community",
        type=str,
        default="public",
        help="community string, default is 'public'",
    )
    parser_v1.set_defaults(snmpinfo=v1info)

    parser_v2c = subparsers.add_parser("v2c", help="Use SNMPv2c")
    parser_v2c.add_argument(
        "-c",
        "--community",
        type=str,
        default="public",
        help="community string, default is 'public'",
    )
    parser_v2c.set_defaults(snmpinfo=v2cinfo)

    parser_v3 = subparsers.add_parser(
        "v3", help="Use SNMPv3 with optional SHA authentication and AES encryption"
    )
    parser_v3.add_argument("secname", type=str, help="SNMPv3 security user name")
    parser_v3.add_argument(
        "--auth-key", type=str, help="authentication key, enables authentication"
    )
    parser_v3.add_argument(
        "--priv-key", type=str, help="private key, requires --auth-key to be set"
    )
    parser_v3.add_argument(
        "--auth-proto",
        type=str,
        choices=["SHA", "MD5"],
        default="SHA",
        help="authentication protocol, requires --auth-key to be set"
    )
    parser_v3.add_argument(
        "--priv-proto",
        type=str,
        choices=["AES", "3DES"],
        default="AES",
        help="privacy protocol, requires --priv-key to be set"
    )
    parser_v3.set_defaults(snmpinfo=v3info)

    group_iprange = parser.add_argument_group("iprange", "Ip range")
    group_iprange.add_argument("--ip-range", type=str, help="IP range for scan")

    group_iplist = parser.add_argument_group("iplist", "Ip list file")
    group_iplist.add_argument("--ip-list-file", type=str, help="IP list file for scan")

    group_seed = parser.add_argument_group("seedrouter", "Discovery using seed router")
    group_seed.add_argument(
        "--seedrouter-address",
        metavar="address",
        type=ipaddress_parser,
        help="the address to start the discovery from",
    )

    parser.add_argument(
        "--exclusion-list",
        metavar="{JSON list}",
        type=exclusion_list_parser,
        help="List of IP addresses, ranges and subnets to exclude",
    )

    args = parser.parse_args()

    valueKeys = ["address", "neighbour"]
    relation_output_stream = csv.DictWriter(
        sys.stdout, fieldnames=valueKeys, lineterminator="\n"
    )

    counter = 0
    if hasattr(args, "snmpinfo"):
        snmp_backend = SNMPBackend(args.snmpinfo(args),
                                   exclusion_list=args.exclusion_list)
    else:
        snmp_backend = None

    if args.seedrouter_address:
        if snmp_backend:
            relation_output_stream.writeheader()
            for link in link_discover_seed(snmp_backend, args.seedrouter_address):
                counter += 1

                relation_output_stream.writerow(
                    {
                        "address": str(link.address_a) if link.address_a else "",
                        "neighbour": str(link.address_b) if link.address_b else "",
                    }
                )
        else:
            logger.warning("No snmp backend for seed router discovery")

    elif args.ip_range or args.ip_list_file:
        # ping to all devices first

        if args.ip_range:
            icmp_backend = FpingBackend(ScanMethod.ip_range, args.exclusion_list)
            live_devices = device_discover_range(icmp_backend, args.ip_range)
        else:
            icmp_backend = FpingBackend(ScanMethod.ip_list, args.exclusion_list)
            live_devices = device_discover_list(icmp_backend, args.ip_list_file)

        device_value_key = ["address"]
        device_output_stream = csv.DictWriter(
            sys.stdout, fieldnames=device_value_key, lineterminator="\n"
        )
        device_output_stream.writeheader()
        for device in live_devices:
            counter += 1
            device_output_stream.writerow({"address": str(device)})

        # use other protocols to scan the IP list before trying snmp
        # scan to discover the links between devices
        if snmp_backend:
            relation_output_stream.writeheader()
            added_links = set()
            for device in live_devices:
                for link in link_discover_single(snmp_backend, device):
                    if link not in added_links:
                        relation_output_stream.writerow(
                            {
                                "address": str(link.address_a)
                                if link.address_a
                                else "",
                                "neighbour": str(link.address_b)
                                if link.address_b
                                else "",
                            }
                        )
                        added_links.add(link)

    else:
        parser.print_help()

    if counter == 0:
        sys.exit(1)


if __name__ == "__main__":
    main()
