import itertools
import logging
import subprocess

from ipaddress import ip_address, IPv4Address, IPv4Network
from typing import Generator, List

from pyautodiscover.consts import ScanMethod
from pyautodiscover.backend import ScanBackendInterface
from pyautodiscover.utils import IPAddress

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class FpingBackend(ScanBackendInterface):
    """
    Run icmp scans for ranges or list of ranges using fping(8).

    Args:
        scan_method (pyautodiscover.consts.ScanMethod): ScanMethod enum member.
        exclusion_list (list): List of addresses/ranges to exclude from results.
        timeout (int): Timeout in seconds for running fping commands.
        fping_path (str): Absolute path to fping binary. By default it's
            assumed that it can be found in PATH.
    """

    def __init__(self, scan_method, exclusion_list=None, timeout=600, fping_path=None):
        self.scan_method = scan_method
        self.exclusion_list = exclusion_list
        self.timeout = timeout
        self.fping_path = fping_path or "fping"

    def scan(self, scan_input: str) -> List[IPv4Address]:
        """
        For a range scan, `scan_input` is an IP range specification, e.g.:
            * 192.0.2.0/24
            * 192.0.2.100-150
            * 172.20,22.10-12,20-22.0/24

        For a list scan, ranges are read from a file named `scan_input` where
        each line can be what range scan supports or a FQDN.

        Returns:
            A sorted list of :py:class:`ipaddress.IPv4Address` objects for the
            nodes that are alive.
        """
        node_iterators = []
        if self.scan_method == ScanMethod.ip_list:
            for range_spec in IPAddress.lines_from_file(scan_input):
                node_iterators += self._scan_range(range_spec)
        elif self.scan_method == ScanMethod.ip_range:
            node_iterators += self._scan_range(scan_input)
        return sorted(itertools.chain.from_iterable(node_iterators))

    def _scan_range(self, range_spec):
        ret = []
        if "/" in range_spec:
            # Handle net/mask, possibly with X-Y ranges, by running one
            # scan per expanded network range.
            for network in IPAddress.ipv4_cidr_range_to_network_list(range_spec):
                ret.append(self._scan_single_network(network))
        else:
            # Without /mask notation, input can generate arbitrary lists of
            # addresses. Run fping with list of IP-addresses as input on stdin.
            if IPAddress.is_fqdn(range_spec):
                input_data = range_spec
            else:
                input_data = "\n".join(
                    str(n) for n in IPAddress.ip_range_to_list(range_spec)
                )
            out = self._fping(input=input_data, timeout=self.timeout)
            ret.append(self._filter_nodes(out))
        return ret

    def _scan_single_network(
        self, network: IPv4Network
    ) -> Generator[IPv4Address, None, None]:
        """Run fping with a network range.

        Returns:
            Nodes alive and not in exclusion list.
        """
        out = self._fping("--generate", str(network), timeout=self.timeout)
        return self._filter_nodes(out)

    def _filter_nodes(self, data: str):
        """Yield IPv4Address/IPv6Address objects of the alive hosts in
        `data`, except those matching the exclusion_list."""
        for line in data.splitlines():
            try:
                addr = ip_address(line)
                if not IPAddress.on_ip_list(line, self.exclusion_list):
                    yield addr
            except ValueError as exc:
                logger.warning("%s", exc)

    def _fping(self, *args, **run_kwargs):
        args = [
            self.fping_path,
            "--quiet",
            "--alive",
            "--addr",
            "--stats",
            *args,
        ]
        logger.debug("Run %s", " ".join(args))
        proc = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="UTF-8",
            **run_kwargs,
        )
        logger.debug(
            "fping returncode: %d, stats/stderr:\n%s", proc.returncode, proc.stderr
        )
        if proc.returncode > 2:
            logger.error(
                "fping returncode %d, stdout: %s", proc.returncode, proc.stdout
            )
            logger.error("fping stderr:\n%s", proc.stderr)
            raise RuntimeError("Failed to scan addresses")
        return proc.stdout
