# -*- coding: utf-8 -*-

from ipaddress import IPv4Address
from collections import namedtuple
from pyautodiscover.backend import (
    CrawlBackendInterface,
    ScanBackendInterface,
    InfoBackendInterface,
)
from pyautodiscover.utils.IPAddress import (
    validate_address_input,
    validate_list_input_file,
)

import logging


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Link(namedtuple("Link", ["address_a", "address_b"])):
    def __hash__(self):
        return hash(frozenset((self.address_a, self.address_b)))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __lt__(self, other):
        return hash(self) < hash(other)


def info(backend, address):
    """
    Retrieves information regarding a device and returns it as a dict of values
    """

    if not isinstance(address, IPv4Address):
        raise TypeError("Expected address to be instance of IPv4Address")

    if not isinstance(backend, InfoBackendInterface):
        raise TypeError("Expected backend to be instance of InfoBackendInterface")

    try:
        return backend.info(address)
    except RuntimeError:
        # there was an error querying for info from the targeted device
        # log it and continue
        logger.warning("Failed to get info for %s", address)
        return {}


def link_discover_seed(backend, base_address):
    """
    Runs discovery from the addresses given and yields Node discovered that
    contains address and parent as IPv4Address objects
    """

    if not isinstance(base_address, IPv4Address):
        raise TypeError("Expected base_address to be instance of IPv4Address")

    if not isinstance(backend, CrawlBackendInterface):
        raise TypeError("Expected backend to be instance of CrawlBackendInterface")

    yielded_links = set()
    checked = set()
    addresses_to_check = [base_address]

    for address in addresses_to_check:
        if address not in checked:
            checked.add(address)
            try:
                for neighbour in backend.neighbours(address):
                    new_link = Link(neighbour, address)
                    if new_link not in yielded_links:
                        yielded_links.add(new_link)
                        addresses_to_check.append(neighbour)
                        yield new_link
            except RuntimeError:
                # some other machine told us that this IP exists so this error
                # might be due to something else. Let's just ignore this error
                # and keep checking the machines that we find.
                logger.warning("Failed to get neighbours of %s", address)


def link_discover_single(backend, address):
    """
    Runs discovery on a address and find all links
    :param backend:
    :param address:
    :return: a list of links
    """

    try:
        for neighbour in backend.neighbours(address):
            new_link = Link(neighbour, address)
            yield new_link
    except RuntimeError:
        # some other machine told us that this IP exists so this error
        # might be due to something else. Let's just ignore this error
        # and keep checking the machines that we find.
        logger.warning("Failed to get neighbours of %s", address)


def device_discover_range(backend, ip_range):
    """
    Run IP range scan to discover devices
    """

    if not validate_address_input(ip_range):
        raise TypeError("Unexpected IP range")
    if not isinstance(backend, ScanBackendInterface):
        raise TypeError("Expected backend to be instance of ScanBackendInterface")

    return backend.scan(ip_range)


def device_discover_list(backend, ip_list_file):
    """
    Run IP list scan to discover devices
    """

    if not isinstance(backend, ScanBackendInterface):
        raise TypeError("Expected backend to be instance of ScanBackendInterface")
    if not validate_list_input_file(ip_list_file):
        raise TypeError("input file: {} contains invalid address".format(ip_list_file))

    return backend.scan(ip_list_file)
