import logging
import re

from ipaddress import ip_address, ip_network
from itertools import product


FQDN_REGEX = re.compile(
    r"^((?!-)[-A-Z\d]{1,62}(?<!-)\.)+[A-Z]{1,62}\.?$", re.IGNORECASE
)
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def generate_sub_range(sub_range):
    """Expand an IPv4 octet string like to a list.

    Example:
        >>> generate_sub_range("0,21-25,252-254")
        [0, 21, 22, 23, 24, 25, 252, 253, 254]
        >>> generate_sub_range("-2,253-,100")
        [0, 1, 2, 253, 254, 255, 100]
        >>> generate_sub_range("100")
        [100]

    """
    if not sub_range:
        raise ValueError("Empty field in IP range")
    elems = sub_range.split(",")
    ret = []
    for elem in elems:
        if "-" in elem:
            a, b = elem.split("-")  # Raises ValueError if length != 2
            a = 0 if a == "" else int(a)
            b = 255 if b == "" else int(b)
            if a > b or b > 255:
                raise ValueError("Bad sub-range")
            ret.extend(range(a, b + 1))
        else:
            elem_val = int(elem)
            if elem_val > 255:
                raise ValueError("Value too high")
            else:
                ret.append(elem_val)

    return ret


def ip_range_to_list(ip_range):
    """Expand ip_range and return a generator of IPv4Address objects."""
    if " " in ip_range:
        return (ip_address(item) for item in ip_range.split(" "))

    sub_ranges = (generate_sub_range(sub_range) for sub_range in ip_range.split("."))
    return (ip_address(bytes(address)) for address in product(*sub_ranges))


def ipv4_cidr_range_to_network_list(range_spec):
    """
    Example:
        >>> gen = ipv4_cidr_range_to_network_list("10.1.20-21,25.128/25")
        >>> next(gen) == ip_network("10.1.20.128/25")
        True
        >>> next(gen) == ip_network("10.1.21.128/25")
        True
        >>> next(gen) == ip_network("10.1.25.128/25")
        True
    """
    range_str, netmask_in = range_spec.split("/")
    try:
        netmask = int(netmask_in)  # CIDR notation, e.g. /26
    except ValueError:
        netmask = str(netmask_in)  # dotted decimal, e.g. 255.255.255.192
    else:
        if not (8 <= netmask <= 32):
            raise ValueError("Invalid netmask length: %s" % netmask)

    sub_ranges = (generate_sub_range(sub_range) for sub_range in range_str.split("."))
    for octets in product(*sub_ranges):
        yield ip_network((bytes(octets), netmask), strict=False)


def is_fqdn(address_input):
    return bool(FQDN_REGEX.match(address_input))


def validate_address_input_range(address_input):

    if is_fqdn(address_input):
        return True
    if "." in address_input:
        if ":" in address_input:
            return False

        ip_fields = address_input.split(".")
        if len(ip_fields) != 4:
            return False
        for field in ip_fields:
            try:
                generate_sub_range(field)
            except ValueError:
                return False
        return True

    elif ":" in address_input:
        ip_fields = address_input.split(":")
        if len(ip_fields) > 8:
            return False
        else:
            for field in ip_fields:
                if field.count("-") >= 2:
                    return False
            return True
    else:
        return False


def validate_address_input(address_input):

    mal_symbols = ";&>$~"
    for c in mal_symbols:
        if c in address_input:
            return False

    # Magellan might use this function to validate a space separated line of
    # input, since a ip_list in magellan can have such lines.
    if " " in address_input:
        for ip in address_input.split(" "):
            if ip:
                ret = validate_address_input(ip)
                if not ret:
                    return False
        return True

    # in case of subnet
    if "/" in address_input:
        if "-" in address_input or "," in address_input:
            try:
                list(ipv4_cidr_range_to_network_list(address_input))
            except ValueError:
                return False
            return True
        try:
            ip_network(address_input, strict=False)
            return True
        except ValueError:
            return False

    # in case of ip range
    return validate_address_input_range(address_input)


def is_address_within_range(address, ip_range):
    if "/" in ip_range:
        return ip_address(address) in ip_network(ip_range, strict=False)

    address_fields = address.split(".")
    range_fields = (generate_sub_range(sub_range) for sub_range in ip_range.split("."))
    for address_field, range_field in zip(address_fields, range_fields):
        if int(address_field) not in range_field:
            return False
    return True


def lines_from_file(filename):
    """Yield IP range specifications read from filename."""
    with open(filename) as fp:
        for line in fp:
            line = line.strip()
            if "#" in line:
                line = line[: line.find("#")]
            if line:
                yield from line.split()


def validate_list_input_file(filename):
    """Return True if filename doesn't contain any lines that are invalid IP
    range specifications and at least one valid. Comments are ignored."""
    found = False
    try:
        for line in lines_from_file(filename):
            found = True
            if not validate_address_input(line):
                logger.error("Invalid range in %s: %r", filename, line)
                return False
    except OSError:
        logger.exception("Failed to read ip list file")
        return False
    return found


def on_ip_list(address, ip_list):
    """ Check whether address is covered by any range in ip_list. """
    return ip_list and any(
        is_address_within_range(address, excluded) for excluded in ip_list
    )
